# -*- coding: utf-8 -*-
# pylint: disable=invalid-name
# pylint: disable=import-outside-toplevel
#
# faucet documentation build configuration file, created by
# sphinx-quickstart on Thu Oct 26 13:48:25 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#

"""
Faucet's Sphinx configuration
"""

import os
import sys

sys.path.insert(0, os.path.abspath("../"))

autodoc_default_options = {
    "members": True,
    "show-inheritance": True,
    "undoc-members": True,
}

# -- Project information -----------------------------------------------------

project = "Faucet"
copyright = "2018-2024, Faucet Developers"  # pylint: disable=redefined-builtin
author = "Faucet Developers"

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
needs_sphinx = "1.8"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.githubpages",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinxcontrib.rsvgconverter",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "README.rst", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
html_sidebars = {
    "**": [
        "relations.html",  # needs 'show_related': True theme option to display
        "searchbox.html",
    ]
}


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "faucetdoc"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "faucet.tex", "Faucet Documentation", "Faucet Developers", "manual"),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "faucet", "Faucet Documentation", [author], 1)]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "faucet",
        "Faucet Documentation",
        author,
        "faucet",
        "",
        "Miscellaneous",
    ),
]

# -- Magic to run sphinx-apidoc automatically -----------------------------

# See https://github.com/rtfd/readthedocs.org/issues/1139
# on which this is based.


def run_apidoc(_):
    """Call sphinx-apidoc on faucet module"""

    from sphinx.ext.apidoc import main as apidoc_main

    apidoc_main(["-e", "-o", "source/apidoc", "../faucet"])


def generate_prometheus_metric_table(_):
    """Autogen prometheus metrics documentation"""

    import faucet.faucet_metrics
    import faucet.gauge_prom
    from prometheus_client import CollectorRegistry

    block_text = {}
    output_path = {
        "faucet": "autogen/faucet_prometheus_metric_table.rst",
        "gauge": "autogen/gauge_prometheus_metric_table.rst",
    }

    metrics = {
        "faucet": faucet.faucet_metrics.FaucetMetrics(reg=CollectorRegistry()),
        "gauge": faucet.gauge_prom.GaugePrometheusClient(reg=CollectorRegistry()),
    }

    for module in ["faucet", "gauge"]:
        block_text[module] = (
            """\
.. list-table:: {} prometheus metrics
    :widths: 40 10 55
    :header-rows: 1

    * - Metric
      - Type
      - Description
""".format(
                module.title()
            )
        )

        # pylint: disable=protected-access
        for metric in metrics[module]._reg.collect():
            if metric.type == "counter":
                metric_name = "{}_total".format(metric.name)
            else:
                metric_name = metric.name

            block_text[
                module
            ] += """\
    * - {}
      - {}
      - {}
""".format(
                metric_name, metric.type, metric.documentation
            )

        with open(output_path[module], "w", encoding="utf-8") as output_file:
            output_file.write(block_text[module])


def setup(app):
    """Add hooks into Sphinx to change behaviour and autogen documentation"""

    # Add custom css
    app.add_css_file("css/responsive-tables.css")
    # Override Sphinx setup to trigger sphinx-apidoc.
    app.connect("builder-inited", run_apidoc)
    app.connect("builder-inited", generate_prometheus_metric_table)



"""Manage flooding/learning on stacked datapaths."""

# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from faucet import valve_of
from faucet.valve_switch_standalone import ValveSwitchManager
from faucet.vlan import NullVLAN


class ValveSwitchStackManagerBase(ValveSwitchManager):
    """Base class for dataplane based flooding/learning on stacked dataplanes."""

    # By default, no reflection used for flooding algorithms.
    _USES_REFLECTION = False

    def __init__(self, stack_manager, **kwargs):
        super().__init__(**kwargs)

        self.stack_manager = stack_manager

        self._set_ext_port_flag = ()
        self._set_nonext_port_flag = ()
        self.external_root_only = False
        if self.has_externals:
            self.logger.info("external ports present, using loop protection")
            self._set_ext_port_flag = (
                self.flood_table.set_external_forwarding_requested(),
            )
            self._set_nonext_port_flag = (
                self.flood_table.set_no_external_forwarding_requested(),
            )
            if (
                not self.stack_manager.stack.is_root()
                and self.stack_manager.stack.is_root_candidate()
            ):
                self.logger.info("external flooding on root only")
                self.external_root_only = True

    @staticmethod
    def _non_stack_learned(other_valves, pkt_meta):
        """
        Obtain DP that has learnt the host that sent the packet

        Args:
            other_valves (list): Other valves
            pkt_meta (PacketMeta): Packet meta sent by the host
        Returns:
            DP: DP that has learnt the host
        """
        other_local_dp_entries = []
        other_external_dp_entries = []
        vlan_vid = pkt_meta.vlan.vid
        for other_valve in other_valves:
            other_dp_vlan = other_valve.dp.vlans.get(vlan_vid, None)
            if other_dp_vlan is not None:
                entry = other_dp_vlan.cached_host(pkt_meta.eth_src)
                if not entry:
                    continue
                if not entry.port.non_stack_forwarding():
                    continue
                if entry.port.loop_protect_external:
                    other_external_dp_entries.append(other_valve.dp)
                else:
                    other_local_dp_entries.append(other_valve.dp)
        # Another DP has learned locally, has priority.
        if other_local_dp_entries:
            return other_local_dp_entries[0]
        # No other DP has learned locally, but at least one has learned externally.
        if other_external_dp_entries:
            entry = pkt_meta.vlan.cached_host(pkt_meta.eth_src)
            # This DP has not learned the host either, use other's external.
            if entry is None:
                return other_external_dp_entries[0]
        return None

    def _external_forwarding_requested(self, port):
        external_forwarding_requested = None
        if self.has_externals:
            if port.tagged_vlans and port.loop_protect_external:
                external_forwarding_requested = False
            elif not port.stack:
                external_forwarding_requested = True
        return external_forwarding_requested

    def _build_flood_acts_for_port(
        self,
        vlan,
        exclude_unicast,
        port,
        exclude_all_external=False,
        exclude_restricted_bcast_arpnd=False,
    ):  # pylint: disable=too-many-arguments,too-many-positional-arguments
        if self.external_root_only:
            exclude_all_external = True
        return super()._build_flood_acts_for_port(
            vlan,
            exclude_unicast,
            port,
            exclude_all_external=exclude_all_external,
            exclude_restricted_bcast_arpnd=exclude_restricted_bcast_arpnd,
        )

    def _flood_actions(
        self,
        in_port,
        external_ports,
        away_flood_actions,
        toward_flood_actions,
        local_flood_actions,
    ):  # pylint: disable=too-many-arguments,too-many-positional-arguments
        raise NotImplementedError

    def _build_flood_rule_actions(
        self,
        vlan,
        exclude_unicast,
        in_port,
        exclude_all_external=False,
        exclude_restricted_bcast_arpnd=False,
    ):  # pylint: disable=too-many-arguments
        """Compiles all the possible flood rule actions for a port on a stack node"""
        exclude_ports = list(self.stack_manager.inactive_away_ports)
        external_ports = vlan.loop_protect_external_ports()
        if in_port and self.stack_manager.is_stack_port(in_port):
            in_port_peer_dp = in_port.stack["dp"]
            exclude_ports = exclude_ports + self.stack_manager.adjacent_stack_ports(
                in_port_peer_dp
            )
        local_flood_actions = tuple(
            self._build_flood_local_rule_actions(
                vlan,
                exclude_unicast,
                in_port,
                exclude_all_external,
                exclude_restricted_bcast_arpnd,
            )
        )
        away_flood_actions = tuple(
            valve_of.flood_tagged_port_outputs(
                self.stack_manager.away_ports, in_port, exclude_ports=exclude_ports
            )
        )
        toward_flood_actions = tuple(
            valve_of.flood_tagged_port_outputs(
                self.stack_manager.chosen_towards_ports, in_port
            )
        )
        flood_acts = self._flood_actions(
            in_port,
            external_ports,
            away_flood_actions,
            toward_flood_actions,
            local_flood_actions,
        )
        return flood_acts

    def _build_mask_flood_rules_filters(
        self, port, vlan, eth_dst, eth_dst_mask, prune
    ):  # pylint: disable=too-many-arguments,too-many-positional-arguments
        """Builds filter for the input table to filter packets on ports that are pruned"""
        ofmsgs = []

        match = {"in_port": port.number, "vlan": vlan}
        if eth_dst is not None:
            match.update({"eth_dst": eth_dst, "eth_dst_mask": eth_dst_mask})

        replace_priority_offset = self.classification_offset - (
            self.pipeline.filter_priority - self.pipeline.select_priority
        )

        priority_offset = replace_priority_offset
        if eth_dst is None:
            priority_offset -= 1

        if prune:
            # Allow the prune rule to be replaced with OF strict matching if
            # this port is unpruned later.
            ofmsgs.extend(
                self.pipeline.filter_packets(match, priority_offset=priority_offset)
            )
        else:
            ofmsgs.extend(
                self.pipeline.remove_filter(match, priority_offset=priority_offset)
            )
            # Control learning from multicast/broadcast on non-root DPs.
            if (
                not self.stack_manager.stack.is_root()
                and eth_dst is not None
                and self._USES_REFLECTION
            ):
                # If this is an edge DP, we don't have to learn from
                # hosts that only broadcast.  If we're an intermediate
                # DP, only learn from broadcasts further away from
                # the root (and ignore the reflected broadcast for
                # learning purposes).
                if (
                    self.stack_manager.stack.is_edge()
                    or self.stack_manager.is_towards_root(port)
                ):
                    ofmsgs.extend(
                        self.pipeline.select_packets(
                            self.flood_table,
                            match,
                            priority_offset=self.classification_offset,
                        )
                    )
        return ofmsgs

    def _build_mask_flood_rules_flood_acts(
        self,
        vlan,
        eth_type,
        eth_dst,
        eth_dst_mask,
        exclude_unicast,
        exclude_restricted_bcast_arpnd,
        command,
        cold_start,
        prune,
        port,
    ):  # pylint: disable=unused-argument,too-many-arguments,too-many-positional-arguments
        """Builds the flood rules for the flood table to forward packets along the stack topology"""
        ofmsgs = []
        flood_acts = []
        if self.has_externals:
            # If external flag is set, flood to external ports, otherwise exclude them.
            for ext_port_flag, exclude_all_external in (
                (valve_of.PCP_NONEXT_PORT_FLAG, True),
                (valve_of.PCP_EXT_PORT_FLAG, False),
            ):
                if not prune:
                    flood_acts, _, _ = self._build_flood_acts_for_port(
                        vlan,
                        exclude_unicast,
                        port,
                        exclude_all_external=exclude_all_external,
                        exclude_restricted_bcast_arpnd=exclude_restricted_bcast_arpnd,
                    )
                port_flood_ofmsg = self._build_flood_rule_for_port(
                    vlan,
                    eth_type,
                    eth_dst,
                    eth_dst_mask,
                    command,
                    port,
                    flood_acts,
                    add_match={valve_of.EXTERNAL_FORWARDING_FIELD: ext_port_flag},
                )
                ofmsgs.append(port_flood_ofmsg)
        else:
            if not prune:
                flood_acts, _, _ = self._build_flood_acts_for_port(
                    vlan,
                    exclude_unicast,
                    port,
                    exclude_restricted_bcast_arpnd=exclude_restricted_bcast_arpnd,
                )
            port_flood_ofmsg = self._build_flood_rule_for_port(
                vlan, eth_type, eth_dst, eth_dst_mask, command, port, flood_acts
            )
            ofmsgs.append(port_flood_ofmsg)
        return ofmsgs

    def _build_mask_flood_rules(
        self,
        vlan,
        eth_type,
        eth_dst,
        eth_dst_mask,
        exclude_unicast,
        exclude_restricted_bcast_arpnd,
        command,
        cold_start,
    ):  # pylint: disable=too-many-arguments,too-many-positional-arguments
        """Builds that flood rules for each mask for each port in the stack.
        This takes into account the pruned and non-pruned ports and returns
            the appropriate flood rule actions"""
        # Stack ports aren't in VLANs, so need special rules to cause flooding from them.
        ofmsgs = super()._build_mask_flood_rules(
            vlan,
            eth_type,
            eth_dst,
            eth_dst_mask,
            exclude_unicast,
            exclude_restricted_bcast_arpnd,
            command,
            cold_start,
        )

        for port in self.stack_manager.stack_ports():
            if eth_dst is not None:
                # Prune broadcast flooding where multiply connected to same DP
                prune = self.stack_manager.is_pruned_port(port)
            else:
                # Do not prune unicast, may be reply from directly connected DP.
                prune = False

            ofmsgs.extend(
                self._build_mask_flood_rules_filters(
                    port, vlan, eth_dst, eth_dst_mask, prune
                )
            )
            ofmsgs.extend(
                self._build_mask_flood_rules_flood_acts(
                    vlan,
                    eth_type,
                    eth_dst,
                    eth_dst_mask,
                    exclude_unicast,
                    exclude_restricted_bcast_arpnd,
                    command,
                    cold_start,
                    prune,
                    port,
                )
            )

        return ofmsgs

    def edge_learn_port(self, other_valves, pkt_meta):
        """
        Find a port towards the edge DP where the packet originated from

        Args:
            other_valves (list): All Valves other than this one.
            pkt_meta (PacketMeta): PacketMeta instance for packet received.
        Returns:
            port to learn host on, or None.
        """
        # Got a packet from another DP.
        if pkt_meta.port.stack:
            # Received packet from
            edge_dp = self._edge_dp_for_host(other_valves, pkt_meta)
            if edge_dp:
                return self.stack_manager.edge_learn_port_towards(pkt_meta, edge_dp)
            # Assuming no DP has learned this host.
            return None

        # Got a packet locally.
        # If learning on an external port, check another DP hasn't
        # already learned on a local/non-external port.
        if pkt_meta.port.loop_protect_external:
            edge_dp = self._non_stack_learned(other_valves, pkt_meta)
            if edge_dp:
                return self.stack_manager.edge_learn_port_towards(pkt_meta, edge_dp)
        # Locally learn.
        return super().edge_learn_port(other_valves, pkt_meta)

    def _edge_dp_for_host(self, other_valves, pkt_meta):
        """Simple distributed unicast learning.

        Args:
            other_valves (list): All Valves other than this one.
            pkt_meta (PacketMeta): PacketMeta instance for packet received.
        Returns:
            Valve instance or None (of edge datapath where packet received)
        """
        raise NotImplementedError

    def add_drop_spoofed_faucet_mac_rules(self, vlan):
        """Install rules to drop spoofed faucet mac"""
        # antispoof for FAUCET's MAC address
        # TODO: antispoof for controller IPs on this VLAN, too.
        ofmsgs = []
        if self.drop_spoofed_faucet_mac:
            for port in self.ports.values():
                if not port.stack:
                    ofmsgs.extend(
                        self.pipeline.filter_packets(
                            {"eth_src": vlan.faucet_mac, "in_port": port.number}
                        )
                    )
        return ofmsgs

    def add_port(self, port):
        ofmsgs = super().add_port(port)
        # If this is a stacking port, accept all VLANs (came from another FAUCET)
        if port.stack:
            # Actual stack traffic will have VLAN tags.
            ofmsgs.append(
                self.vlan_table.flowdrop(
                    match=self.vlan_table.match(in_port=port.number, vlan=NullVLAN()),
                    priority=self.low_priority + 1,
                )
            )
            ofmsgs.append(
                self.vlan_table.flowmod(
                    match=self.vlan_table.match(in_port=port.number),
                    priority=self.low_priority,
                    inst=self.pipeline.accept_to_classification(),
                )
            )
        return ofmsgs

    def del_port(self, port):
        ofmsgs = super().del_port(port)
        if port.stack:
            for vlan in self.vlans.values():
                vlan.clear_cache_hosts_on_port(port)
            ofmsgs.extend(self._del_host_flows(port))
        return ofmsgs

    def get_lacp_dpid_nomination(self, lacp_id, valve, other_valves):
        """Chooses the DP for a given LAG.

        The DP will be nominated by the following conditions in order:
            1) Number of LAG ports
            2) Root DP
            3) Lowest DPID

        Args:
            lacp_id: The LACP LAG ID
            other_valves (list): list of other valves
        Returns:
            nominated_dpid, reason
        """
        if not other_valves:
            return None, ""
        stacked_other_valves = self.stack_manager.stacked_valves(other_valves)
        all_stacked_valves = {valve}.union(stacked_other_valves)
        ports = {}
        no_sync_ports = {}
        root_dpid = None
        for stack_valve in all_stacked_valves:
            all_lags = stack_valve.dp.lags_up()
            if lacp_id in all_lags:
                ports[stack_valve.dp.dp_id] = len(all_lags[lacp_id])
            nosync_lags = stack_valve.dp.lags_nosync()
            for l_id in nosync_lags:
                ports.setdefault(stack_valve.dp.dp_id, 0)
                no_sync_ports[stack_valve.dp.dp_id] = len(nosync_lags.get(l_id, 0))
            if stack_valve.dp.stack.is_root():
                root_dpid = stack_valve.dp.dp_id
        # Order by number of ports
        port_order = sorted(
            ports,
            key=lambda port: (ports.get(port, 0), no_sync_ports.get(port, 0)),
            reverse=True,
        )
        if not port_order:
            return None, ""
        most_ports_dpid = port_order[0]
        most_ports_dpids = [
            dpid for dpid, num in ports.items() if num == ports[most_ports_dpid]
        ]
        if len(most_ports_dpids) > 1:
            # There are several dpids that have the same number of lags
            if root_dpid in most_ports_dpids:
                # root_dpid is the chosen DPID
                return root_dpid, "root dp"
            # Order by lowest DPID
            return sorted(most_ports_dpids), "lowest dpid"
        # Most_ports_dpid is the chosen DPID
        return most_ports_dpid, "most LAG ports"

    def _learn_host_intervlan_routing_flows(self, port, vlan, eth_src, eth_dst):
        """Returns flows for the eth_src_table that enable packets that have been
           routed to be accepted from an adjacent DP and then switched to the destination.
           Eth_src_table flow rule to match on port, eth_src, eth_dst and vlan

        Args:
            port (Port): Port to match on.
            vlan (VLAN): VLAN to match on
            eth_src: source MAC address (should be the router MAC)
            eth_dst: destination MAC address
        """
        ofmsgs = []
        (src_rule_idle_timeout, src_rule_hard_timeout, _) = self._learn_host_timeouts(
            port, eth_src
        )
        src_match = self.eth_src_table.match(
            vlan=vlan, eth_src=eth_src, eth_dst=eth_dst
        )
        src_priority = self.host_priority - 1
        inst = (self.eth_src_table.goto(self.output_table),)
        ofmsgs.extend(
            [
                self.eth_src_table.flowmod(
                    match=src_match,
                    priority=src_priority,
                    inst=inst,
                    hard_timeout=src_rule_hard_timeout,
                    idle_timeout=src_rule_idle_timeout,
                )
            ]
        )
        return ofmsgs

    def _valve_learn_host_from_pkt(self, valve, now, pkt_meta, other_valves):
        """Add L3 forwarding rule if necessary for inter-VLAN routing."""
        ofmsgs_by_valve = super().learn_host_from_pkt(
            valve, now, pkt_meta, other_valves
        )
        if (
            self.stack_manager.stack.route_learning
            and not self.stack_manager.stack.is_root()
        ):
            if pkt_meta.eth_src == pkt_meta.vlan.faucet_mac:
                ofmsgs_by_valve[valve].extend(
                    self._learn_host_intervlan_routing_flows(
                        pkt_meta.port, pkt_meta.vlan, pkt_meta.eth_src, pkt_meta.eth_dst
                    )
                )
            elif pkt_meta.eth_dst == pkt_meta.vlan.faucet_mac:
                ofmsgs_by_valve[valve].extend(
                    self._learn_host_intervlan_routing_flows(
                        pkt_meta.port, pkt_meta.vlan, pkt_meta.eth_dst, pkt_meta.eth_src
                    )
                )
        return ofmsgs_by_valve

    def learn_host_from_pkt(self, valve, now, pkt_meta, other_valves):
        ofmsgs_by_valve = {}

        if self.stack_manager.stack.route_learning:
            stacked_other_valves = self.stack_manager.stacked_valves(other_valves)
            all_stacked_valves = {valve}.union(stacked_other_valves)

            # NOTE: multi DP routing requires learning from directly attached switch first.
            if pkt_meta.port.stack:
                peer_dp = pkt_meta.port.stack["dp"]
                if peer_dp.dyn_running:
                    faucet_macs = {pkt_meta.vlan.faucet_mac}.union(
                        {valve.dp.faucet_dp_mac for valve in all_stacked_valves}
                    )
                    # Must always learn FAUCET VIP, but rely on neighbor
                    # to learn other hosts first.
                    if pkt_meta.eth_src not in faucet_macs:
                        return ofmsgs_by_valve

            for other_valve in stacked_other_valves:
                stack_port = other_valve.stack_manager.relative_port_towards(
                    self.stack_manager.stack.name
                )
                valve_vlan = other_valve.dp.vlans.get(pkt_meta.vlan.vid, None)
                if stack_port and valve_vlan:
                    valve_pkt_meta = copy.copy(pkt_meta)
                    valve_pkt_meta.vlan = valve_vlan
                    valve_pkt_meta.port = stack_port
                    valve_other_valves = all_stacked_valves - {other_valve}
                    ofmsgs_by_valve.update(
                        self._valve_learn_host_from_pkt(
                            other_valve, now, valve_pkt_meta, valve_other_valves
                        )
                    )

        ofmsgs_by_valve.update(
            self._valve_learn_host_from_pkt(valve, now, pkt_meta, other_valves)
        )
        return ofmsgs_by_valve


class ValveSwitchStackManagerNoReflection(ValveSwitchStackManagerBase):
    """Stacks of size 2 - all switches directly connected to root.

    Root switch simply floods to all other switches.

    Non-root switches simply flood to the root.
    """

    def _flood_actions(
        self,
        in_port,
        external_ports,
        away_flood_actions,
        toward_flood_actions,
        local_flood_actions,
    ):  # pylint: disable=too-many-arguments,too-many-positional-arguments
        if not in_port or self.stack_manager.is_stack_port(in_port):
            flood_prefix = ()
        else:
            if external_ports:
                flood_prefix = self._set_nonext_port_flag
            else:
                flood_prefix = self._set_ext_port_flag

        flood_actions = (
            flood_prefix
            + toward_flood_actions
            + away_flood_actions
            + local_flood_actions
        )

        return flood_actions

    def _edge_dp_for_host(self, other_valves, pkt_meta):
        """Size 2 means root shortest path is always directly connected."""
        peer_dp = pkt_meta.port.stack["dp"]
        if peer_dp.dyn_running:
            return self._non_stack_learned(other_valves, pkt_meta)
        # Fall back to assuming peer knows if we are not the peer's controller.
        return peer_dp


class ValveSwitchStackManagerReflection(ValveSwitchStackManagerBase):
    """Stacks size > 2 reflect floods off of root (selective flooding).

    .. code-block:: none

                             Hosts
                              ||||
                              ||||
                +----+       +----+       +----+
             ---+1   |       |1234|       |   1+---
       Hosts ---+2   |       |    |       |   2+--- Hosts
             ---+3   |       |    |       |   3+---
             ---+4  5+-------+5  6+-------+5  4+---
                +----+       +----+       +----+

                Root DP

    Non-root switches flood only to the root. The root switch
    reflects incoming floods back out. Non-root switches flood
    packets from the root locally and to switches further away
    from the root. Flooding is entirely implemented in the dataplane.

    A host connected to a non-root switch can receive a copy of its
    own flooded packet (because the non-root switch does not know
    it has seen the packet already).

    A host connected to the root switch does not have this problem
    (because flooding is always away from the root). Therefore,
    connections to other non-FAUCET stacking networks should only
    be made to the root.

    On the root switch (left), flood destinations are:

    1: 2 3 4 5(s)
    2: 1 3 4 5(s)
    3: 1 2 4 5(s)
    4: 1 2 3 5(s)
    5: 1 2 3 4 5(s, note reflection)

    On the middle switch:

    1: 5(s)
    2: 5(s)
    3: 5(s)
    4: 5(s)
    5: 1 2 3 4 6(s)
    6: 5(s)

    On the rightmost switch:

    1: 5(s)
    2: 5(s)
    3: 5(s)
    4: 5(s)
    5: 1 2 3 4
    """

    # Indicate to base class use of reflection required.
    _USES_REFLECTION = True

    def _learn_cache_check(
        self,
        entry,
        vlan,
        now,
        eth_src,
        port,
        ofmsgs,
        cache_port,
        cache_age,
        delete_existing,
        refresh_rules,
    ):  # pylint: disable=too-many-arguments,too-many-positional-arguments
        learn_exit = False
        update_cache = True
        if cache_port is not None:
            # packet was received on same member of a LAG.
            same_lag = port.lacp and port.lacp == cache_port.lacp
            # stacks of size > 2 will have an unknown MAC flooded towards the root,
            # and flooded down again. If we learned the MAC on a local port and
            # heard the reflected flooded copy, discard the reflection.
            local_stack_learn = port.stack and not cache_port.stack
            guard_time = self.cache_update_guard_time
            if cache_port == port or same_lag or local_stack_learn:
                port_cache_valid = (
                    port.dyn_update_time is not None
                    and port.dyn_update_time <= entry.cache_time
                )
                # aggressively re-learn on LAGs, and prefer recently learned
                # locally learned hosts on a stack.
                if same_lag or local_stack_learn:
                    guard_time = 2
                # port didn't change status, and recent cache update, don't do anything.
                if cache_age < guard_time and port_cache_valid:
                    update_cache = False
                    learn_exit = True
                # skip delete if host didn't change ports or on same LAG.
                elif cache_port == port or same_lag:
                    delete_existing = False
                    if port_cache_valid:
                        refresh_rules = True
        return (
            learn_exit,
            ofmsgs,
            cache_port,
            update_cache,
            delete_existing,
            refresh_rules,
        )

    def _flood_actions(
        self,
        in_port,
        external_ports,
        away_flood_actions,
        toward_flood_actions,
        local_flood_actions,
    ):  # pylint: disable=too-many-arguments,too-many-positional-arguments
        if self.stack_manager.stack.is_root():
            if external_ports:
                flood_prefix = self._set_nonext_port_flag
            else:
                flood_prefix = self._set_ext_port_flag
            flood_actions = away_flood_actions + local_flood_actions

            if in_port and self.stack_manager.is_away(in_port):
                # Packet from a non-root switch, flood locally and to all non-root switches
                # (reflect it).
                flood_actions = (
                    away_flood_actions
                    + (valve_of.output_in_port(),)
                    + local_flood_actions
                )

            flood_actions = flood_prefix + flood_actions
        else:
            # Default non-root strategy is flood towards root.
            if external_ports:
                flood_actions = self._set_nonext_port_flag + toward_flood_actions
            else:
                flood_actions = self._set_ext_port_flag + toward_flood_actions

            if in_port:
                # Packet from switch further away, flood it to the root.
                if self.stack_manager.is_away(in_port):
                    flood_actions = toward_flood_actions
                # Packet from the root.
                elif self.stack_manager.is_towards_root(in_port):
                    # If we have external ports, and packet hasn't already been flooded
                    # externally, flood it externally before passing it to further away switches,
                    # and mark it flooded.
                    if external_ports:
                        flood_actions = (
                            self._set_nonext_port_flag
                            + away_flood_actions
                            + local_flood_actions
                        )
                    else:
                        flood_actions = (
                            away_flood_actions
                            + self._set_nonext_port_flag
                            + local_flood_actions
                        )
                # Packet from external port, locally. Mark it already flooded externally and
                # flood to root (it came from an external switch so keep it within the stack).
                elif in_port.loop_protect_external:
                    flood_actions = self._set_nonext_port_flag + toward_flood_actions
                else:
                    flood_actions = self._set_ext_port_flag + toward_flood_actions

        return flood_actions

    def _edge_dp_for_host(self, other_valves, pkt_meta):
        """For stacks size > 2."""
        # TODO: currently requires controller to manage all switches
        # in the stack to keep each DP's graph consistent.
        # TODO: simplest possible unicast learning.
        # We find just one port that is the shortest unicast path to
        # the destination. We could use other factors (eg we could
        # load balance over multiple ports based on destination MAC).
        # Find port that forwards closer to destination DP that
        # has already learned this host (if any).
        peer_dp = pkt_meta.port.stack["dp"]
        if peer_dp.dyn_running:
            return self._non_stack_learned(other_valves, pkt_meta)
        # Fall back to peer knows if edge or root if we are not the peer's controller.
        if peer_dp.stack.is_edge() or peer_dp.stack.is_root():
            return peer_dp
        # No DP has learned this host, yet. Take no action to allow remote learning to occur.
        return None



"""Standard FAUCET pipeline."""

# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from faucet.faucet_metadata import EGRESS_METADATA_MASK


class ValveTableConfig:  # pylint: disable=too-many-instance-attributes
    """Configuration for a single table."""

    def __init__(
        self,
        name,
        table_id,
        exact_match=None,
        meter=None,
        output=True,
        miss_goto=None,
        size=None,
        match_types=None,
        set_fields=None,
        dec_ttl=None,
        vlan_scale=None,
        vlan_port_scale=None,
        next_tables=None,
        metadata_match=0,
        metadata_write=0,
    ):  # pylint: disable=too-many-arguments,too-many-positional-arguments
        self.name = name
        self.table_id = table_id
        self.exact_match = exact_match
        self.meter = meter
        self.output = output
        self.miss_goto = miss_goto
        self.size = size
        self.match_types = match_types
        self.set_fields = set_fields
        self.dec_ttl = dec_ttl
        self.vlan_scale = vlan_scale
        self.vlan_port_scale = vlan_port_scale
        self.metadata_match = metadata_match
        self.metadata_write = metadata_write
        if next_tables:
            assert isinstance(next_tables, (list, tuple))
            self.next_tables = next_tables
        else:
            self.next_tables = ()

    def __str__(self):
        field_strs = " ".join(
            ["%s: %s" % (key, val) for key, val in sorted(self.__dict__.items()) if val]
        )
        return "table config %s" % field_strs

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return hash(self.__str__())

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()

    def __lt__(self, other):
        return self.__hash__() < other.__hash__()


_NEXT_ETH = ("eth_dst_hairpin", "eth_dst", "flood")
_NEXT_VIP = ("vip",) + _NEXT_ETH


def _fib_table(ipv, table_id):
    return ValveTableConfig(
        "ipv%u_fib" % ipv,
        table_id,
        match_types=(
            ("eth_type", False),
            ("ipv%u_dst" % ipv, True),
            ("vlan_vid", False),
        ),
        set_fields=("eth_dst", "eth_src", "vlan_vid"),
        dec_ttl=True,
        vlan_port_scale=3.1,
        next_tables=_NEXT_VIP,
    )


PORT_ACL_DEFAULT_CONFIG = ValveTableConfig(
    "port_acl",
    0,
    match_types=(("in_port", False),),
    next_tables=(("vlan",) + _NEXT_VIP),
)
VLAN_DEFAULT_CONFIG = ValveTableConfig(
    "vlan",
    PORT_ACL_DEFAULT_CONFIG.table_id + 1,
    match_types=(
        ("eth_dst", True),
        ("eth_type", False),
        ("in_port", False),
        ("vlan_vid", False),
    ),
    set_fields=("vlan_vid",),
    vlan_port_scale=3,
    next_tables=("copro", "vlan_acl", "classification", "eth_src"),
)
COPRO_DEFAULT_CONFIG = ValveTableConfig(
    "copro",
    VLAN_DEFAULT_CONFIG.table_id + 1,
    match_types=(("in_port", False), ("eth_type", False), ("vlan_vid", False)),
    vlan_port_scale=1.5,
    miss_goto="eth_dst",
    next_tables=(("eth_dst",)),
)
VLAN_ACL_DEFAULT_CONFIG = ValveTableConfig(
    "vlan_acl",
    VLAN_DEFAULT_CONFIG.table_id + 1,
    next_tables=(("classification", "eth_src") + _NEXT_ETH),
)
CLASSIFICATION_DEFAULT_CONFIG = ValveTableConfig(
    "classification",
    VLAN_ACL_DEFAULT_CONFIG.table_id + 1,
    miss_goto="eth_src",
    next_tables=(("eth_src", "ipv4_fib", "ipv6_fib") + _NEXT_VIP),
)
ETH_SRC_DEFAULT_CONFIG = ValveTableConfig(
    "eth_src",
    CLASSIFICATION_DEFAULT_CONFIG.table_id + 1,
    miss_goto="eth_dst",
    next_tables=(("ipv4_fib", "ipv6_fib") + _NEXT_VIP),
    match_types=(
        ("eth_dst", True),
        ("eth_src", False),
        ("eth_type", False),
        ("in_port", False),
        ("vlan_vid", False),
    ),
    set_fields=("vlan_vid", "eth_dst"),
    vlan_port_scale=4.1,
)
IPV4_FIB_DEFAULT_CONFIG = _fib_table(4, ETH_SRC_DEFAULT_CONFIG.table_id + 1)
IPV6_FIB_DEFAULT_CONFIG = _fib_table(6, IPV4_FIB_DEFAULT_CONFIG.table_id + 1)
VIP_DEFAULT_CONFIG = ValveTableConfig(
    "vip",
    IPV6_FIB_DEFAULT_CONFIG.table_id + 1,
    match_types=(
        ("arp_tpa", False),
        ("eth_dst", False),
        ("eth_type", False),
        ("icmpv6_type", False),
        ("ip_proto", False),
    ),
    next_tables=_NEXT_ETH,
    vlan_scale=8,
)
ETH_DST_HAIRPIN_DEFAULT_CONFIG = ValveTableConfig(
    "eth_dst_hairpin",
    VIP_DEFAULT_CONFIG.table_id + 1,
    match_types=(("in_port", False), ("eth_dst", False), ("vlan_vid", False)),
    miss_goto="eth_dst",
    exact_match=True,
    vlan_port_scale=4.1,
)
ETH_DST_DEFAULT_CONFIG = ValveTableConfig(
    "eth_dst",
    ETH_DST_HAIRPIN_DEFAULT_CONFIG.table_id + 1,
    exact_match=True,
    miss_goto="flood",  # Note: when using egress acls the miss goto will be
    # egress acl table
    match_types=(("eth_dst", False), ("vlan_vid", False)),
    next_tables=("egress", "egress_acl"),
    vlan_port_scale=4.1,
    metadata_write=EGRESS_METADATA_MASK,
)
EGRESS_ACL_DEFAULT_CONFIG = ValveTableConfig(
    "egress_acl", ETH_DST_DEFAULT_CONFIG.table_id + 1, next_tables=("egress",)
)
EGRESS_DEFAULT_CONFIG = ValveTableConfig(
    "egress",
    EGRESS_ACL_DEFAULT_CONFIG.table_id + 1,
    match_types=(("metadata", True), ("vlan_vid", False)),
    vlan_port_scale=1.5,
    next_tables=("flood",),
    miss_goto="flood",
    metadata_match=EGRESS_METADATA_MASK,
)
FLOOD_DEFAULT_CONFIG = ValveTableConfig(
    "flood",
    EGRESS_DEFAULT_CONFIG.table_id + 1,
    match_types=(("eth_dst", True), ("in_port", False), ("vlan_vid", False)),
    vlan_port_scale=8.0,
)
MINIMUM_FAUCET_PIPELINE_TABLES = {"vlan", "eth_src", "eth_dst", "flood"}

# TODO: implement an eth_type table before VLAN. This would enable interception
# of control protocols and simplify matches in vlan/eth_src, enabling use of
# exact_match.
FAUCET_PIPELINE = (
    PORT_ACL_DEFAULT_CONFIG,
    VLAN_DEFAULT_CONFIG,
    COPRO_DEFAULT_CONFIG,
    VLAN_ACL_DEFAULT_CONFIG,
    CLASSIFICATION_DEFAULT_CONFIG,
    ETH_SRC_DEFAULT_CONFIG,
    IPV4_FIB_DEFAULT_CONFIG,
    IPV6_FIB_DEFAULT_CONFIG,
    VIP_DEFAULT_CONFIG,
    ETH_DST_HAIRPIN_DEFAULT_CONFIG,
    ETH_DST_DEFAULT_CONFIG,
    EGRESS_ACL_DEFAULT_CONFIG,
    EGRESS_DEFAULT_CONFIG,
    FLOOD_DEFAULT_CONFIG,
)

DEFAULT_CONFIGS = {
    "port_acl": PORT_ACL_DEFAULT_CONFIG,
    "vlan": VLAN_DEFAULT_CONFIG,
    "copro": COPRO_DEFAULT_CONFIG,
    "vlan_acl": VLAN_ACL_DEFAULT_CONFIG,
    "eth_src": ETH_SRC_DEFAULT_CONFIG,
    "ipv4_fib": IPV4_FIB_DEFAULT_CONFIG,
    "ipv6_fib": IPV6_FIB_DEFAULT_CONFIG,
    "vip": VIP_DEFAULT_CONFIG,
    "eth_dst_hairpin": ETH_DST_HAIRPIN_DEFAULT_CONFIG,
    "eth_dst": ETH_DST_DEFAULT_CONFIG,
    "egress_acl": EGRESS_ACL_DEFAULT_CONFIG,
    "egress": EGRESS_DEFAULT_CONFIG,
    "flood": FLOOD_DEFAULT_CONFIG,
}



"""This module contains code relating to the use of OpenFlow Metadata within
Faucet.
"""

PORT_METADATA_MASK = 0xFFF
VLAN_METADATA_MASK = 0xFFF000
EGRESS_METADATA_MASK = PORT_METADATA_MASK | VLAN_METADATA_MASK


def get_egress_metadata(port_num, vid):
    """Return the metadata value to output a packet to port port_num on vlan
    vid"""
    metadata = vid << 12 | (port_num & PORT_METADATA_MASK)
    return metadata, EGRESS_METADATA_MASK



"""Base class for generative testing framework"""

import random
import unittest
import networkx

from mininet.log import error
from clib.mininet_test_watcher import OptimizedTopologyWatcher
from clib.mininet_test_base_topo import FaucetTopoTestBase
from clib.mininet_test_topo import FAUCET


class FaucetFaultToleranceBaseTest(FaucetTopoTestBase):
    """
    Generate a topology of the given parameters (using build_net & TopoBaseTest)
        and then call network function to test the network and then slowly tear out bits
        until the expected host connectivity does not match the real host connectivity.
    ===============================================================================================
    INSTANT_FAIL:
        The fault-tolerant tests will continue fail if there is a pair of hosts that can not
            establish a connection.
        Set to true to allow the test suite to continue testing the connectivity
            for a fault to build the full graph for the current fault.
    ASSUME_SYMMETRIC_PING:
        A simplification can assume that (h1 -> h2) implies (h2 -> h1).
        Set to true to assume that host connectivity is symmetric.
    INTERVLAN_ONLY:
        Set to true to test only the inter-VLAN connectivity; ignore connections between hosts on
            the same VLAN. Speed up the inter-VLAN testing by ignoring the intra-VLAN cases for
            tests that inherit from a intra-VLAN test. This creates that assumption that inter-VLAN
            does not disrupt the intra-VLAN.
    ASSUME_SYMMETRIC_PING:
        A simplification can assume that (h1 -> h2) implies (h2 -> h1).
        Set to true to assume that host connectivity is symmetric.
        Symmetric only assumption produces `\\sum_{i=1}^{#nodes} i` pings per stage
    ASSUME_TRANSITIVE_PING:
        Assume for (h1 -> h2) & (h2 -> h3) then (h1 -> h3)
        Set to true to assume that host connectivity is transitive
        Transitive only assumption produces `#nodes` pings per stage
    ===============================================================================================
    """

    INSTANT_FAIL = True
    INTERVLAN_ONLY = False
    ASSUME_SYMMETRIC_PING = True
    ASSUME_TRANSITIVE_PING = True

    # Watches the faults and host connectvitiy
    topo_watcher = None

    # Randomization variables
    seed = 1
    rng = None

    # List of fault events
    fault_events = None
    # Number of faults to occur before recalculating connectivity
    num_faults = 1

    # Fault-tolerance tests will only work in software
    SOFTWARE_ONLY = True

    # Number of Faucet controllers to create
    NUM_FAUCET_CONTROLLERS = 2
    # Number of VLANs to create, if >= 2 then routing will be applied
    NUM_VLANS = None
    # Number of DPs in the network
    NUM_DPS = None
    # Number of links between switches
    N_DP_LINKS = None

    # Dictionary of roots to stack priority value
    stack_roots = None

    host_links = None
    switch_links = None
    routers = None

    def setUp(self):
        pass

    def set_up(self, network_graph, stack_roots, host_links=None, host_vlans=None):
        """
        Args:
            network_graph (networkx.MultiGraph): Network topology for the test
            stack_roots (dict): The priority values for the stack roots
            host_links (dict): Links for each host to switches
            host_vlans (dict): VLAN for each host
        """
        super().setUp()
        switch_links = list(network_graph.edges()) * self.N_DP_LINKS
        link_vlans = {edge: None for edge in switch_links}
        if not host_links or not host_vlans:
            # Setup normal host links & vlans
            host_links = {}
            host_vlans = {}
            host_n = 0
            for dp_i in network_graph.nodes():
                for v_i in range(self.NUM_VLANS):
                    host_links[host_n] = [dp_i]
                    host_vlans[host_n] = v_i
                    host_n += 1
        dp_options = {}
        for i in network_graph.nodes():
            dp_options.setdefault(
                i,
                {
                    "group_table": self.GROUP_TABLE,
                    "ofchannel_log": (
                        self.debug_log_path + str(i) if self.debug_log_path else None
                    ),
                    "hardware": "Open vSwitch",
                },
            )
            if i in stack_roots:
                dp_options[i]["stack"] = {"priority": stack_roots[i]}
        vlan_options = {}
        routers = {}
        if self.NUM_VLANS >= 2:
            # Setup options for routing
            routers = {0: list(range(self.NUM_VLANS))}
            for i in range(self.NUM_VLANS):
                vlan_options[i] = {
                    "faucet_mac": self.faucet_mac(i),
                    "faucet_vips": [self.faucet_vip(i)],
                    "targeted_gw_resolution": False,
                }
            for i in network_graph.nodes():
                dp_options[i]["arp_neighbor_timeout"] = 2
                dp_options[i]["max_resolve_backoff_time"] = 2
                dp_options[i]["proactive_learn_v4"] = True
        self.host_links = host_links
        self.switch_links = switch_links
        self.routers = routers
        self.stack_roots = stack_roots
        self.build_net(
            host_links=host_links,
            host_vlans=host_vlans,
            switch_links=switch_links,
            link_vlans=link_vlans,
            n_vlans=self.NUM_VLANS,
            dp_options=dp_options,
            vlan_options=vlan_options,
            routers=routers,
        )
        self.start_net()

    def host_connectivity(self, host, dst):
        """Ping to a destination, return True if the ping was successful"""
        try:
            self._ip_ping(host, dst, 5, timeout=50, count=5, require_host_learned=False)
        except AssertionError:
            return False
        return True

    def calculate_connectivity(self):
        """Ping between each set of host pairs to calculate host connectivity"""
        connected_hosts = self.topo_watcher.get_connected_hosts(
            symmetric=self.ASSUME_SYMMETRIC_PING,
            transitive=self.ASSUME_TRANSITIVE_PING,
            intervlan_only=self.INTERVLAN_ONLY,
        )
        actual_graph = networkx.MultiDiGraph()
        for src, dst in connected_hosts.edges():
            src_id = self.topo.nodeInfo(src)["host_n"]
            dst_id = self.topo.nodeInfo(dst)["host_n"]
            result = self.host_connectivity(
                self.host_information[src_id]["host"],
                self.host_information[dst_id]["ip"].ip,
            )
            if result:
                actual_graph.add_edge(src, dst)
            if self.INSTANT_FAIL:
                self.assertTrue(result, "Connection failed: %s -/-> %s" % (src, dst))
        self.assertEqual(
            list(connected_hosts.edges()),
            list(actual_graph.edges()),
            "Resulting host connectivity graph does not match expected (%s != %s)"
            % (list(connected_hosts.edges()), list(actual_graph.edges())),
        )

    def create_controller_fault(self, *args):
        """
        Set controller down (disconnects all switches from the controller)
        Args:
            index: The index to the controller to take down
        """
        index = args[0]
        controller = self.net.controllers[index]
        controller.stop()
        self.net.controllers.remove(controller)
        self.topo_watcher.add_fault("Controller %s DOWN" % controller.name)

    def get_faucet_controllers(self):
        """Return list of Faucet controllers"""
        return [c for c in self.net.controllers if isinstance(c, FAUCET)]

    def create_random_controller_fault(self, *_args):
        """Randomly create a fault for a controller"""
        controllers = self.get_faucet_controllers()
        if len(controllers) == 1:
            return
        i = random.randrange(len(controllers))
        c_name = controllers[i].name
        controller = next(
            (cont for cont in self.net.controllers if cont.name == c_name), None
        )
        if controller is None:
            return
        self.create_controller_fault(self.net.controllers.index(controller))

    def create_switch_fault(self, *args):
        """
        Set switch down (Deletes the OVS switch bridge)
        Args:
            index: Index of the switch dpid to take out
        """
        index = args[0]
        dpid = self.dpids[index]
        switch_name = self.topo.switches_by_id[index]
        switch = next(
            (switch for switch in self.net.switches if switch.name == switch_name), None
        )
        if switch is None:
            return
        self.dump_switch_flows(switch)
        name = "%s:%s DOWN" % (self.topo.switches_by_id[index], self.dpids[index])
        self.topo_watcher.add_switch_fault(index, name)
        switch.stop()
        error(switch.cmd(self.VSCTL, "del-controller", switch.name, "|| true"))
        self.assertTrue(
            self.wait_for_prometheus_var(
                "dp_status", 0, default=0, dpid=dpid, retries=10
            ),
            "DP %s not detected as DOWN" % dpid,
        )
        self.net.switches.remove(switch)

    def random_switch_fault(self, *_args):
        """Randomly take out an available switch"""
        sw_list = self.topo_watcher.get_eligable_switch_events()
        index_list = []
        for sw_name in sw_list:
            index_list.append(self.topo.nodeInfo(sw_name)["switch_n"])
        if len(self.stack_roots.keys()) <= 1:
            # Prevent the only root from being destroyed
            sorted_roots = dict(
                sorted(self.stack_roots.items(), key=lambda item: item[1])
            )
            for root_index in sorted_roots.keys():
                if root_index in index_list:
                    index_list.remove(root_index)
        if not index_list:
            return
        index = self.rng.randrange(len(index_list))
        sw_index = index_list[index]
        self.create_switch_fault(sw_index)

    def dp_link_fault(self, *args):
        """
        Create a fault/tear down the stack link between two switches
        Args:
            src_dp_index: Index of the source DP of the stack link
            dst_dp_index: Index of the destination DP of the stack
        """
        src_i = args[0]
        dst_i = args[1]
        src_dpid = self.dpids[src_i]
        dst_dpid = self.dpids[dst_i]
        s1_name = self.topo.switches_by_id[src_i]
        s2_name = self.topo.switches_by_id[dst_i]
        for port, link in self.topo.ports[s1_name].items():
            status = self.stack_port_status(src_dpid, s1_name, port)
            if link[0] == s2_name and status == 3:
                peer_port = link[1]
                self.set_port_down(port, src_dpid)
                self.set_port_down(peer_port, dst_dpid)
                self.wait_for_stack_port_status(src_dpid, s1_name, port, 4)
                self.wait_for_stack_port_status(dst_dpid, s2_name, peer_port, 4)
                name = "Link %s[%s]:%s-%s[%s]:%s DOWN" % (
                    s1_name,
                    src_dpid,
                    port,
                    s2_name,
                    dst_dpid,
                    peer_port,
                )
                self.topo_watcher.add_link_fault(src_i, dst_i, name)
                return

    def random_link_fault(self, *_args):
        """Randomly create a fault for a DP link"""
        link_list = self.topo_watcher.get_eligable_link_events()
        if not link_list:
            return
        index = self.rng.randrange(len(link_list))
        dp_link = link_list[index]
        src_i = self.topo.nodeInfo(dp_link[0])["switch_n"]
        dst_i = self.topo.nodeInfo(dp_link[1])["switch_n"]
        self.dp_link_fault(src_i, dst_i)

    def create_proportional_random_fault_event(self):
        """Create a fault-event randomly based on the number of link and switch events available"""
        funcs = []
        for _ in self.topo_watcher.get_eligable_link_events():
            funcs.append(self.random_link_fault)
        for _ in self.topo_watcher.get_eligable_switch_events():
            funcs.append(self.random_switch_fault)
        for _ in range(len(self.get_faucet_controllers()) - 1):
            funcs.append(self.create_random_controller_fault)
        random.shuffle(funcs)
        i = self.rng.randrange(len(funcs))
        funcs[i]()

    def network_function(self, fault_events=None, num_faults=1):
        """
        Test the network by slowly tearing it down different ways
        Args:
            fault_events: (optional) list of tuples of fault event functions and the parameters to
                use in the given order; instead of randomly choosing parts of the network to break
            num_faults: (optional) number of faults to cause before each evaluation is made
        """
        self.verify_stack_up()

        self.fault_events = fault_events
        self.num_faults = num_faults
        self.rng = random.Random(self.seed)

        self.topo_watcher = OptimizedTopologyWatcher(
            self.topo, self.host_information, self.configuration_options["routers"]
        )

        # Calculate stats (before any tear downs)
        self.calculate_connectivity()
        # Start tearing down the network
        if self.fault_events:
            # Do Specified list of faults (in order) until failure or fault list completed
            fault_index = 0
            while fault_index < len(self.fault_events):
                for _ in range(self.num_faults):
                    event_func, params = self.fault_events[fault_index]
                    fault_index += 1
                    event_func(*params)
                    self.calculate_connectivity()
        else:
            # Continue creating fault until none are available or expected connectivity does not
            #      match real connectivity
            while self.topo_watcher.continue_faults() or bool(
                len(self.get_faucet_controllers()) - 1
            ):
                for _ in range(self.num_faults):
                    self.create_proportional_random_fault_event()
                self.calculate_connectivity()

    def tearDown(self, ignore_oferrors=False):
        """Make sure to dump the watcher information too"""
        if self.topo_watcher:
            self.topo_watcher.dump_info(self.tmpdir)
        super().tearDown(ignore_oferrors=ignore_oferrors)


class FaucetFaultTolerance2DPTest(FaucetFaultToleranceBaseTest):
    """Run a range of fault-tolerance tests for topologies on 2 DPs"""

    NUM_DPS = 2
    NUM_HOSTS = 4
    NUM_VLANS = 2
    N_DP_LINKS = 1
    STACK_ROOTS = {0: 1}
    ASSUME_SYMMETRIC_PING = False


class FaucetFaultTolerance3DPTest(FaucetFaultToleranceBaseTest):
    """Run a range of fault-tolerance tests for topologies on 3 DPs"""

    NUM_DPS = 3
    NUM_HOSTS = 6
    NUM_VLANS = 2
    N_DP_LINKS = 1
    STACK_ROOTS = {0: 1}


class FaucetFaultTolerance4DPTest(FaucetFaultToleranceBaseTest):
    """Run a range of fault-tolerance tests for topologies on 4 DPs"""

    NUM_DPS = 4
    NUM_HOSTS = 4
    NUM_VLANS = 1
    N_DP_LINKS = 1
    STACK_ROOTS = {0: 1}

    def test_ftp2_all_random_switch_failures(self):
        """Test fat-tree-pod-2 randomly tearing down only switches"""
        fault_events = [
            (self.random_switch_fault, (None,)) for _ in range(self.NUM_DPS)
        ]
        stack_roots = {2 * i: 1 for i in range(self.NUM_DPS // 2)}
        self.set_up(networkx.cycle_graph(self.NUM_DPS), stack_roots)
        self.network_function(fault_events=fault_events)

    def test_ftp2_all_random_link_failures(self):
        """Test fat-tree-pod-2 randomly tearing down only switch-switch links"""
        network_graph = networkx.cycle_graph(self.NUM_DPS)
        fault_events = [
            (self.random_link_fault, (None,)) for _ in range(len(network_graph.edges()))
        ]
        stack_roots = {2 * i: 1 for i in range(self.NUM_DPS // 2)}
        self.set_up(network_graph, stack_roots)
        self.network_function(fault_events=fault_events)

    def test_ftp2_edge_root_link_fault(self):
        """Test breaking a link between a edge switch to the root aggregation switch"""
        fault_events = [(self.dp_link_fault, (0, 3))]
        stack_roots = {2 * i: i + 1 for i in range(self.NUM_DPS // 2)}
        self.set_up(networkx.cycle_graph(self.NUM_DPS), stack_roots)
        self.network_function(fault_events=fault_events)

    def test_ftp2_destroying_one_of_each_link(self):
        """Test tearing down one of each link for a fat-tree-pod-2 with redundant edges"""
        self.N_DP_LINKS = 2
        fault_events = []
        for i in range(self.NUM_DPS):
            j = i + 1 if i + 1 < self.NUM_DPS else 0
            fault_events.append((self.dp_link_fault, (i, j)))
        num_faults = len(fault_events)
        stack_roots = {2 * i: 1 for i in range(self.NUM_DPS // 2)}
        self.set_up(networkx.cycle_graph(self.NUM_DPS), stack_roots)
        self.network_function(fault_events=fault_events, num_faults=num_faults)
        self.N_DP_LINKS = 1


class FaucetFaultTolerance5DPTest(FaucetFaultToleranceBaseTest):
    """Run a range of fault-tolerance tests for topologies on 5 DPs"""

    NUM_DPS = 5
    NUM_HOSTS = 5
    NUM_VLANS = 1
    N_DP_LINKS = 1
    STACK_ROOTS = {0: 1}


class FaucetFaultTolerance6DPTest(FaucetFaultToleranceBaseTest):
    """Run a range of fault-tolerance tests for topologies on 5 DPs"""

    NUM_DPS = 6
    NUM_HOSTS = 6
    NUM_VLANS = 1
    N_DP_LINKS = 1
    STACK_ROOTS = {0: 1}


@unittest.skip("Too computationally complex")
class FaucetFaultTolerance7DPTest(FaucetFaultToleranceBaseTest):
    """Run a range of fault-tolerance tests for topologies on 5 DPs"""

    NUM_DPS = 7
    NUM_HOSTS = 7
    NUM_VLANS = 1
    N_DP_LINKS = 1
    STACK_ROOTS = {0: 1}


TEST_CLASS_LIST = [
    FaucetFaultTolerance2DPTest,
    FaucetFaultTolerance3DPTest,
    FaucetFaultTolerance4DPTest,
    FaucetFaultTolerance5DPTest,
    FaucetFaultTolerance6DPTest,
    FaucetFaultTolerance7DPTest,
]
MIN_NODES = min([c.NUM_DPS for c in TEST_CLASS_LIST])
MAX_NODES = max([c.NUM_DPS for c in TEST_CLASS_LIST])



"""Configure routing between VLANs."""

# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytricia

from faucet.conf import Conf, test_config_condition


# pylint: disable=c-extension-no-member
class _PyTricia(pytricia.PyTricia):
    def __repr__(self):
        return str([(k, self[k]) for k in sorted(self.keys())])


class Router(Conf):
    """Implement FAUCET configuration for a router."""

    defaults = {
        "bgp": {},
        "vlans": None,
    }

    defaults_types = {
        "bgp": dict,
        "vlans": list,
    }

    ipaddress_fields = ("neighbor_addresses", "server_addresses")

    bgp_defaults_types = {
        "as": int,
        "connect_mode": str,
        "neighbor_addresses": list,
        "neighbor_as": int,
        "port": int,
        "routerid": str,
        "server_addresses": list,
        "vlan": (str, int),
    }

    def __init__(self, _id, dp_id, conf):
        self.bgp = {}
        self.vlans = []
        self.vip_map_by_ipv = {}
        super().__init__(_id, dp_id, conf)

    def _sub_conf_val(self, sub_conf, key):
        try:
            return self.__dict__[sub_conf][key]
        except KeyError:
            return None

    def _bgp_val(self, key):
        return self._sub_conf_val("bgp", key)

    def __str__(self):
        return str(self._id)

    def check_config(self):
        super().check_config()
        if self.bgp:
            self._check_conf_types(self.bgp, self.bgp_defaults_types)
            self.bgp = self._set_unknown_conf(self.bgp, self.bgp_defaults_types)
            if not self.bgp_connect_mode():
                self.bgp["connect_mode"] = "passive"
            for field in self.ipaddress_fields:
                if field in self.bgp:
                    self.bgp[field] = frozenset(
                        [self._check_ip_str(ip_str) for ip_str in self.bgp[field]]
                    )
            for accessor_val, required_field in (
                (self.bgp_ipvs(), "server_addresses"),
                (self.bgp_as(), "as"),
                (self.bgp_port(), "port"),
                (self.bgp_connect_mode(), "connect_mode"),
                (self.bgp_routerid(), "routerid"),
                (self.bgp_neighbor_addresses(), "neighbor_addresses"),
                (self.bgp_neighbor_as(), "neighbor_as"),
            ):
                test_config_condition(
                    not accessor_val, "BGP %s must be specified" % required_field
                )
            test_config_condition(
                self.bgp_connect_mode() != "passive", "BGP connect_mode must be passive"
            )
            for ipv in self.bgp_ipvs():
                test_config_condition(
                    len(self.bgp_server_addresses_by_ipv(ipv)) != 1,
                    "Only one BGP server address per IP version supported",
                )
            if not self.bgp_vlan():
                test_config_condition(
                    len(self.vlans) != 1,
                    "If routing more than one VLAN, must specify BGP VLAN",
                )
                self.set_bgp_vlan(self.vlans[0])
        else:
            test_config_condition(
                not self.vlans,
                "A router must have least one VLAN specified at top level",
            )

    def vip_map(self, ipa):
        """Return VIP for IP address, if any."""
        if ipa.version in self.vip_map_by_ipv:
            result = self.vip_map_by_ipv[ipa.version].get(ipa)
            if result:
                return result
        return (None, None)

    def finalize(self):
        for vlan in self.vlans:
            for faucet_vip in vlan.faucet_vips:
                ipv = faucet_vip.version
                if ipv not in self.vip_map_by_ipv:
                    self.vip_map_by_ipv[ipv] = _PyTricia(faucet_vip.ip.max_prefixlen)
                self.vip_map_by_ipv[ipv][faucet_vip.network] = (vlan, faucet_vip)
        super().finalize()

    def bgp_as(self):
        """Return BGP AS."""
        return self._bgp_val("as")

    def bgp_connect_mode(self):
        """Return BGP connect mode."""
        return self._bgp_val("connect_mode")

    def bgp_neighbor_addresses(self):
        """Return BGP neighbor addresses."""
        return self._bgp_val("neighbor_addresses")

    def bgp_neighbor_as(self):
        """Return BGP neighbor AS number."""
        return self._bgp_val("neighbor_as")

    def bgp_port(self):
        """Return BGP port."""
        return self._bgp_val("port")

    def bgp_routerid(self):
        """Return BGP router ID."""
        return self._bgp_val("routerid")

    def bgp_server_addresses(self):
        """Return BGP server addresses."""
        return self._bgp_val("server_addresses")

    def bgp_vlan(self):
        """Return BGP VLAN."""
        return self._bgp_val("vlan")

    def set_bgp_vlan(self, vlan):
        """Set BGP VLAN."""
        if self.bgp:
            self.bgp["vlan"] = vlan

    def bgp_ipvs(self):
        """Return list of IP versions for BGP configured on this VLAN."""
        return self._ipvs(self.bgp_server_addresses())

    def bgp_neighbor_addresses_by_ipv(self, ipv):
        """Return BGP neighbor addresses with specified IP version on this VLAN."""
        return self._by_ipv(self.bgp_neighbor_addresses(), ipv)

    def bgp_server_addresses_by_ipv(self, ipv):
        """Return BGP server addresses with specified IP version on this VLAN."""
        return self._by_ipv(self.bgp_server_addresses(), ipv)



"""Configuration for a stack."""

# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import Counter
import networkx

from faucet.conf import Conf, test_config_condition


class Stack(Conf):
    """Stores state related to DP stack information, this includes the current elected root as that
    is technically a fixed allocation for this DP Stack instance."""

    defaults = {
        # Sets the root priority value of the current DP with stacking
        "priority": None,
        # Use the stack route algorithms, will be forced true if routing is enabled
        "route_learning": False,
        # Number of update time intervals for a down stack node to still be considered healthy
        "down_time_multiple": 3,
        # Minimum percentage value of required UP stack ports for this stack
        # node to be considered healthy
        "min_stack_health": 1.0,
        # Minimum percentage value of required UP LACP ports for this stack
        # node to be considered healthy
        "min_lacp_health": 1.0,
    }

    defaults_types = {
        "priority": int,
        "route_learning": bool,
        "down_time_multiple": int,
        "min_stack_health": float,
        "min_lacp_health": float,
    }

    def __init__(
        self, _id, dp_id, name, canonical_port_order, lacp_down_ports, lacp_ports, conf
    ):
        """
        Constructs a new stack object

        Args:
            _id (str): Name of the configuration key
            dp_id (int): DP ID of the DP that holds this stack instance
            name (str): Name of the DP that holds this stack instance
            canonical_port_order (func): Function to order ports in a standardized way
            lacp_down_ports (func): Returns a tuple of the not UP LACP ports for this stack node
            lacp_ports (func): Returns a tuple of all LACP ports for this stack node
            conf (dict): Stack configuration
        """
        self.name = name

        # Function to order ports in a standardized way
        self.canonical_port_order = canonical_port_order
        self.lacp_down_ports = lacp_down_ports
        self.lacp_ports = lacp_ports

        # Stack configuration options
        self.priority = None
        self.route_learning = None
        self.down_time_multiple = None
        self.min_stack_health = None
        self.min_lacp_health = None

        # Ports that have stacking configured
        self.ports = []

        # Stack graph containing all the DPs & ports in the stacking topology
        self.graph = None

        # Additional stacking information
        self.root_name = None
        self.roots_names = None
        self.root_flood_reflection = None

        # Whether the stack node is currently healthy
        # dyn_healthy_info := (<running>, <stack ports>, <lacp ports>)
        self.dyn_healthy_info = (False, 0.0, 0.0)
        self.dyn_healthy = False

        super().__init__(_id, dp_id, conf)

    def clone_dyn_state(self, prev_stack, dps=None):
        """Copy dyn state from the old stack instance when warm/cold starting"""
        if prev_stack:
            self.dyn_healthy = prev_stack.dyn_healthy
            self.dyn_healthy_info = prev_stack.dyn_healthy_info
            if dps:
                stack_port_dps = [dp for dp in dps if dp.stack_ports()]
                for dp in stack_port_dps:
                    for port in dp.stack_ports():
                        port_up = False
                        if port.is_stack_up():
                            port_up = True
                        elif port.is_stack_init() and port.stack["port"].is_stack_up():
                            port_up = True
                        self.modify_link(dp, port, add=port_up)

    def live_timeout_healthy(self, last_live_time, now, update_time):
        """
        Determines the timeout of the current stack node, and whether
        the current stack node can be considered healthy according to
        the `down_time_multiple` number of stack root update time intervals.

        Args:
            last_live_time (float): Last known live time for this current stack node
            now (float): Current time
            update_time (int): Update time interval
        Return:
            bool: If node down time is still in update time interval threshold; considered healthy,
            float: Time elapsed since timed out
        """
        # Time elapsed for the number of safe down time multiples before considered unhealthy
        down_time = self.down_time_multiple * update_time
        # Final time at which nodes are still considered healthy
        health_timeout = now - down_time
        # If node last known live time was greater than the health timeout then it is healthy
        timeout_healthy = last_live_time >= health_timeout
        return timeout_healthy, health_timeout

    def stack_port_healthy(self):
        """
        Determines the percentage of UP stack ports, and whether
        the current stack node can be considered healthy according to
        the `min_stack_health` configuration option.

        Return:
            bool: Whether threshold from DOWN stack ports is met; considered healthy,
            float: Percentage of stack ports UP out of all stack ports
        """
        down_ports = self.down_ports()
        all_ports = self.ports
        if len(all_ports) == 0:
            return True, 1.0
        percentage = 1.0 - (float(len(down_ports) / float(len(all_ports))))
        stack_ports_healthy = percentage >= self.min_stack_health
        return stack_ports_healthy, percentage

    def lacp_port_healthy(self):
        """
        Determines the percentage of UP LACP ports, and whether
        the current stack node can be considered healthy according to
        the `min_lacp_health` configuration option.

        Return:
            bool: Whether threshold from DOWN LACP ports is met; considered healthy,
            float: Percentage of LACP ports UP out of all lacp ports
        """
        down_ports = self.lacp_down_ports()
        all_ports = self.lacp_ports()
        if len(all_ports) == 0:
            return True, 1.0
        percentage = 1.0 - (float(len(down_ports) / float(len(all_ports))))
        lacp_ports_healthy = percentage >= self.min_lacp_health
        return lacp_ports_healthy, percentage

    def update_health(self, now, dp_last_live_time, update_time):
        """
        Determines whether the current stack node is healthy

        Args:
            now (float): Current time
            last_live_times (dict): Last live time value for each DP
            update_time (int): Stack root update interval time
        Return:
            tuple: Current stack node health state,
            str: Reason for the current state
        """
        reason = ""
        last_live_time = dp_last_live_time.get(self.name, 0)
        timeout_healthy, health_timeout = self.live_timeout_healthy(
            last_live_time, now, update_time
        )
        if not timeout_healthy:
            # Too long since DP last running, if DP not running then
            #   number of UP stack or LACP ports should be 0
            reason += "last running %us ago (timeout %us)" % (
                now - last_live_time,
                health_timeout,
            )
            self.dyn_healthy_info = (False, 0.0, 0.0)
            self.dyn_healthy = False
            return self.dyn_healthy, reason
        reason += "running %us ago" % (now - last_live_time)
        if reason:
            reason += ", "
        stack_ports_healthy, stack_percentage = self.stack_port_healthy()
        if not stack_ports_healthy:
            # The number of DOWN stack ports surpasses the threshold for DOWN stack port tolerance
            reason += "stack ports %s (%.0f%%) not up" % (
                list(self.down_ports()),
                (1.0 - stack_percentage) * 100.0,
            )
        else:
            reason += "%.0f%% stack ports running" % (stack_percentage * 100.0)
        if self.lacp_ports():
            if reason:
                reason += ", "
            lacp_ports_healthy, lacp_percentage = self.lacp_port_healthy()
            if not lacp_ports_healthy:
                # The number of DOWN LACP ports surpasses the threshold for DOWN LACP port tolerance
                reason += "lacp ports %s (%.0f%%) not up" % (
                    list(self.lacp_down_ports()),
                    (1.0 - lacp_percentage) * 100.0,
                )
            else:
                reason += "%.0f%% lacp ports running" % (lacp_percentage * 100.0)
        else:
            # No LACP ports in node, so default to 100% UP & don't report information
            lacp_ports_healthy = True
            lacp_percentage = 0.0
        self.dyn_healthy_info = (timeout_healthy, stack_percentage, lacp_percentage)
        if timeout_healthy and stack_ports_healthy and lacp_ports_healthy:
            self.dyn_healthy = True
        else:
            self.dyn_healthy = False
        return self.dyn_healthy, reason

    @staticmethod
    def nominate_stack_root(stacks):
        """Return stack names in priority order and the chosen root"""

        def health_priority(stack):
            # Invert the health priority info so it is sorted correctly
            #   in relation to priority and the binary health
            invert_info = (
                1.0 - stack.dyn_healthy_info[1],
                1.0 - stack.dyn_healthy_info[2],
            )
            return (not stack.dyn_healthy, *invert_info, stack.priority, stack.dp_id)

        stack_priorities = sorted(stacks, key=health_priority)
        priority_names = tuple(stack.name for stack in stack_priorities)
        nominated_name = priority_names[0]
        return priority_names, nominated_name

    def resolve_topology(self, dps, meta_dp_state):
        """
        Resolve & verify correct inter-DP stacking config

        Args:
            dps (list): List of configured DPs
            meta_dp_state (MetaDPState): Provided if reloading when choosing a new root DP
        """
        stack_dps = [dp for dp in dps if dp.stack is not None]
        stack_priority_dps = [dp for dp in stack_dps if dp.stack.priority]
        stack_port_dps = [dp for dp in dps if dp.stack_ports()]

        if not stack_priority_dps:
            test_config_condition(stack_dps, "stacking enabled but no root DP")
            return

        if not self.ports:
            return

        for dp in stack_priority_dps:
            test_config_condition(
                not isinstance(dp.stack.priority, int),
                (
                    "stack priority must be type %s not %s"
                    % (int, type(dp.stack.priority))
                ),
            )
            test_config_condition(
                dp.stack.priority <= 0, ("stack priority must be > 0")
            )

        self.roots_names, self.root_name = self.nominate_stack_root(
            [dp.stack for dp in stack_priority_dps]
        )

        if meta_dp_state:
            # If meta_dp_state exists, then we are reloading a new instance of the stack
            #   for a new 'dynamically' chosen root
            if meta_dp_state.stack_root_name in self.roots_names:
                self.root_name = meta_dp_state.stack_root_name

        for dp in stack_port_dps:
            for vlan in dp.vlans.values():
                if vlan.faucet_vips:
                    self.route_learning = True

        edge_count = Counter()
        graph = networkx.MultiGraph()
        for dp in stack_port_dps:
            graph.add_node(dp.name)
            for port in dp.stack_ports():
                edge_name = Stack.modify_topology(graph, dp, port)
                edge_count[edge_name] += 1
        for edge_name, count in edge_count.items():
            test_config_condition(
                count != 2, "%s defined only in one direction" % edge_name
            )
        if graph.size() and self.name in graph:
            self.graph = graph
            for dp in graph.nodes():
                path_to_root_len = len(self.shortest_path(self.root_name, src_dp=dp))
                test_config_condition(
                    path_to_root_len == 0, "%s not connected to stack" % dp
                )
            root_len = self.longest_path_to_root_len()
            if root_len is not None and root_len > 2:
                self.root_flood_reflection = True

    @staticmethod
    def modify_topology(graph, dp, port, add=True):
        """Add/remove an edge to the stack graph which originates from this dp and port."""

        def canonical_edge(dp, port):
            peer_dp = port.stack["dp"]
            peer_port = port.stack["port"]
            sort_edge_a = (dp.name, port.name, dp, port)
            sort_edge_z = (peer_dp.name, peer_port.name, peer_dp, peer_port)
            sorted_edge = sorted((sort_edge_a, sort_edge_z))
            edge_a, edge_b = sorted_edge[0][2:], sorted_edge[1][2:]
            return edge_a, edge_b

        def make_edge_name(edge_a, edge_z):
            edge_a_dp, edge_a_port = edge_a
            edge_z_dp, edge_z_port = edge_z
            return "%s:%s-%s:%s" % (
                edge_a_dp.name,
                edge_a_port.name,
                edge_z_dp.name,
                edge_z_port.name,
            )

        def make_edge_attr(edge_a, edge_z):
            edge_a_dp, edge_a_port = edge_a
            edge_z_dp, edge_z_port = edge_z
            return {
                "dp_a": edge_a_dp,
                "port_a": edge_a_port,
                "dp_z": edge_z_dp,
                "port_z": edge_z_port,
            }

        edge = canonical_edge(dp, port)
        edge_a, edge_z = edge
        edge_name = make_edge_name(edge_a, edge_z)
        edge_attr = make_edge_attr(edge_a, edge_z)
        edge_a_dp, _ = edge_a
        edge_z_dp, _ = edge_z
        if add:
            graph.add_edge(
                edge_a_dp.name, edge_z_dp.name, key=edge_name, port_map=edge_attr
            )
        elif (edge_a_dp.name, edge_z_dp.name, edge_name) in graph.edges:
            graph.remove_edge(edge_a_dp.name, edge_z_dp.name, edge_name)

        return edge_name

    def modify_link(self, dp, port, add=True):
        """Update the stack topology according to the event"""
        return Stack.modify_topology(self.graph, dp, port, add)

    def hash(self):
        """Return hash of a topology graph"""
        return hash(tuple(sorted(self.graph.degree())))

    def get_node_link_data(self):
        """Return network stacking graph as a node link representation"""
        return networkx.readwrite.json_graph.node_link_data(self.graph)

    def add_port(self, port):
        """Add a port to this stack"""
        self.ports.append(port)

    def any_port_up(self):
        """Return true if any stack port is UP"""
        for port in self.ports:
            if port.is_stack_up():
                return True
        return False

    def down_ports(self):
        """Return tuple of not running stack ports"""
        return tuple(port for port in self.ports if not port.is_stack_up())

    def canonical_up_ports(self, ports=None):
        """Obtains list of UP stack ports in canonical order"""
        if ports is None:
            ports = self.ports
        return self.canonical_port_order([port for port in ports if port.is_stack_up()])

    def shortest_path(self, dest_dp, src_dp=None):
        """Return shortest path to a DP, as a list of DPs."""
        if src_dp is None:
            src_dp = self.name
        if self.graph:
            try:
                return sorted(networkx.all_shortest_paths(self.graph, src_dp, dest_dp))[
                    0
                ]
            except (networkx.exception.NetworkXNoPath, networkx.exception.NodeNotFound):
                pass
        return []

    def shortest_path_to_root(self, src_dp=None):
        """Return shortest path to root DP, as list of DPs."""
        return self.shortest_path(self.root_name, src_dp=src_dp)

    def is_root(self):
        """Return True if this DP is the root of the stack."""
        return self.name == self.root_name

    def is_root_candidate(self):
        """Return True if this DP could be a root of the stack."""
        return self.name in self.roots_names

    def is_edge(self):
        """Return True if this DP is a stack edge."""
        return not self.is_root() and self.longest_path_to_root_len() == len(
            self.shortest_path_to_root()
        )

    def shortest_path_port(self, dest_dp):
        """Return first port on our DP, that is the shortest path towards dest DP."""
        shortest_path = self.shortest_path(dest_dp)
        if len(shortest_path) > 1:
            peer_dp = shortest_path[1]
            peer_dp_ports = self.peer_up_ports(peer_dp)
            if peer_dp_ports:
                return peer_dp_ports[0]
        return None

    def peer_up_ports(self, peer_dp):
        """Return list of stack ports that are up towards a peer."""
        return self.canonical_port_order(
            [
                port
                for port in self.ports
                if port.running() and (port.stack["dp"].name == peer_dp)
            ]
        )

    def longest_path_to_root_len(self):
        """Return length of the longest path to root in the stack."""
        if not self.graph or not self.root_name:
            return None
        len_paths_to_root = [
            len(self.shortest_path(self.root_name, src_dp=dp))
            for dp in self.graph.nodes()
        ]
        if len_paths_to_root:
            return max(len_paths_to_root)
        return None

    def is_in_path(self, src_dp, dst_dp):
        """Return True if the current DP is in the path from src_dp to dst_dp

        Args:
            src_dp (str): DP name
            dst_dp (str): DP name
        Returns:
            bool: True if self is in the path from the src_dp to the dst_dp.
        """
        path = self.shortest_path(dst_dp, src_dp=src_dp)
        return self.name in path

    def peer_symmetric_up_ports(self, peer_dp):
        """Return list of stack ports that are up towards us from a peer"""
        # Sort adjacent ports by canonical port order
        return self.canonical_port_order(
            [
                port.stack["port"]
                for port in self.ports
                if port.running() and (port.stack["dp"].name == peer_dp)
            ]
        )

    def shortest_symmetric_path_port(self, peer_dp):
        """Return port on our DP that is the first port of the adjacent DP towards us"""
        shortest_path = self.shortest_path(self.name, src_dp=peer_dp)
        if len(shortest_path) == 2:
            adjacent_up_ports = self.peer_symmetric_up_ports(peer_dp)
            if adjacent_up_ports:
                return adjacent_up_ports[0].stack["port"]
        return None



"""Manage LLDP."""

# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import defaultdict

from faucet import valve_util
from faucet import valve_of
from faucet import valve_packet
from faucet.valve_manager_base import ValveManagerBase


class ValveLLDPManager(ValveManagerBase):
    """Manage LLDP."""

    def __init__(
        self,
        vlan_table,
        highest_priority,
        logger,
        notify,
        inc_var,
        set_var,
        set_port_var,
        stack_manager,
    ):  # pylint: disable=too-many-arguments,too-many-positional-arguments
        self.vlan_table = vlan_table
        self.highest_priority = highest_priority
        self.logger = logger
        self.notify = notify
        self._set_var = set_var
        self._inc_var = inc_var
        self._set_port_var = set_port_var
        self.stack_manager = stack_manager

    def _lldp_match(self, port):
        return self.vlan_table.match(
            in_port=port.number,
            eth_dst=valve_packet.LLDP_MAC_NEAREST_BRIDGE,
            eth_dst_mask=valve_packet.BRIDGE_GROUP_MASK,
            eth_type=valve_of.ether.ETH_TYPE_LLDP,
        )

    def add_port(self, port):
        ofmsgs = []
        if port.receive_lldp:
            ofmsgs.append(
                self.vlan_table.flowcontroller(
                    match=self._lldp_match(port),
                    priority=self.highest_priority,
                    max_len=128,
                )
            )
        return ofmsgs

    def del_port(self, port):
        ofmsgs = []
        if port.receive_lldp:
            ofmsgs.append(
                self.vlan_table.flowdel(
                    match=self._lldp_match(port), priority=self.highest_priority
                )
            )
        return ofmsgs

    def verify_lldp(
        self,
        port,
        now,
        valve,
        other_valves,
        remote_dp_id,
        remote_dp_name,
        remote_port_id,
        remote_port_state,
    ):  # pylint: disable=too-many-arguments,too-many-positional-arguments
        """
        Verify correct LLDP cabling, then update port to next state

        Args:
            port (Port): Port that received the LLDP
            now (float): Current time
            other_valves (list): Other valves in the topology
            remote_dp_id (int): Received LLDP remote DP ID
            remote_dp_name (str): Received LLDP remote DP name
            remote_port_id (int): Recevied LLDP port ID
            remote_port_state (int): Received LLDP port state
        Returns:
            dict: Ofmsgs by valve
        """
        if not port.stack:
            return {}
        remote_dp = port.stack["dp"]
        remote_port = port.stack["port"]
        stack_correct = True
        self._inc_var("stack_probes_received")
        if (
            remote_dp_id != remote_dp.dp_id
            or remote_dp_name != remote_dp.name
            or remote_port_id != remote_port.number
        ):
            self.logger.error(
                "Stack %s cabling incorrect, expected %s:%s:%u, actual %s:%s:%u"
                % (
                    port,
                    valve_util.dpid_log(remote_dp.dp_id),
                    remote_dp.name,
                    remote_port.number,
                    valve_util.dpid_log(remote_dp_id),
                    remote_dp_name,
                    remote_port_id,
                )
            )
            stack_correct = False
            self._inc_var("stack_cabling_errors")
        port.dyn_stack_probe_info = {
            "last_seen_lldp_time": now,
            "stack_correct": stack_correct,
            "remote_dp_id": remote_dp_id,
            "remote_dp_name": remote_dp_name,
            "remote_port_id": remote_port_id,
            "remote_port_state": remote_port_state,
        }
        return self.update_stack_link_state([port], now, valve, other_valves)

    def update_stack_link_state(self, ports, now, valve, other_valves):
        """
        Update the stack link states of the set of provided stack ports

        Args:
            ports (list): List of stack ports to update the state of
            now (float): Current time
            valve (Valve): Valve that owns this LLDPManager instance
            other_valves (list): List of other valves
        Returns:
            dict: ofmsgs by valve
        """
        stack_changes = 0
        ofmsgs_by_valve = defaultdict(list)
        stacked_valves = set()
        if self.stack_manager:
            stacked_valves = {valve}.union(
                self.stack_manager.stacked_valves(other_valves)
            )
        for port in ports:
            before_state = port.stack_state()
            after_state, reason = port.stack_port_update(now)
            if before_state != after_state:
                self._set_port_var("port_stack_state", after_state, port)
                self._inc_var(
                    "port_stack_state_change_count",
                    labels=valve.dp.port_labels(port.number),
                )
                self.notify(
                    {"STACK_STATE": {"port": port.number, "state": after_state}}
                )
                self.logger.info(
                    "Stack %s state %s (previous state %s): %s"
                    % (
                        port,
                        port.stack_state_name(after_state),
                        port.stack_state_name(before_state),
                        reason,
                    )
                )
                stack_changes += 1
                port_up = False
                if port.is_stack_up():
                    port_up = True
                elif port.is_stack_init() and port.stack["port"].is_stack_up():
                    port_up = True
                for stack_valve in stacked_valves:
                    stack_valve.stack_manager.update_stack_topo(port_up, valve.dp, port)
        if stack_changes or valve.stale_root:
            self.logger.info(
                "%u stack ports changed state, stale root %s"
                % (stack_changes, valve.stale_root)
            )
            valve.stale_root = False
            notify_dps = {}
            for stack_valve in stacked_valves:
                if not stack_valve.dp.dyn_running:
                    continue
                ofmsgs_by_valve[stack_valve].extend(
                    stack_valve.add_vlans(stack_valve.dp.vlans.values())
                )
                for port in stack_valve.dp.stack_ports():
                    ofmsgs_by_valve[stack_valve].extend(
                        stack_valve.switch_manager.del_port(port)
                    )
                ofmsgs_by_valve[stack_valve].extend(
                    stack_valve.stack_manager.add_tunnel_acls()
                )
                path_port = stack_valve.stack_manager.chosen_towards_port
                path_port_number = path_port.number if path_port else 0.0
                self._set_var(
                    "dp_root_hop_port",
                    path_port_number,
                    labels=stack_valve.dp.base_prom_labels(),
                )
                notify_dps.setdefault(stack_valve.dp.name, {})[
                    "root_hop_port"
                ] = path_port_number
            # Find the first valve with a valid stack and trigger notification.
            for stack_valve in stacked_valves:
                if stack_valve.dp.stack.graph:
                    self.notify(
                        {
                            "STACK_TOPO_CHANGE": {
                                "stack_root": stack_valve.dp.stack.root_name,
                                "graph": stack_valve.dp.stack.get_node_link_data(),
                                "dps": notify_dps,
                            }
                        }
                    )
                    break
        return ofmsgs_by_valve



"""Manage higher level stack functions"""

from collections import defaultdict

from faucet.valve_manager_base import ValveManagerBase


class ValveStackManager(ValveManagerBase):
    """Implement stack manager, this handles the more higher-order stack functions.
    This includes port nominations and flood directionality."""

    def __init__(
        self, logger, dp, stack, tunnel_acls, acl_manager, output_table, **_kwargs
    ):
        """
        Initialize variables and set up peer distances

        Args:
            stack (Stack): Stack object of the DP on the Valve being managed
        """
        # Logger for logging
        self.logger = logger
        # DP instance for stack healthyness
        self.dp = dp  # pylint: disable=invalid-name
        # Stack instance
        self.stack = stack

        # Used the manage the tunnel ACLs which requires stack knowledge
        self.tunnel_acls = tunnel_acls
        self.acl_manager = acl_manager
        self.output_table = output_table

        # Ports that are the shortest distance to the root
        self.towards_root_ports = None
        # Ports on an adjacent DP that is the chosen shortest path to the root
        self.chosen_towards_ports = None
        # Single port on the adjacent shortest path DP
        self.chosen_towards_port = None

        # All ports that are not the shortest distance to the root
        self.away_ports = None
        # Ports whose peer DPs have a shorter path to root
        self.inactive_away_ports = None
        # Redundant ports for each adjacent DP
        self.pruned_away_ports = None

        self.reset_peer_distances()

    @staticmethod
    def stacked_valves(valves):
        """Return set of valves that have stacking enabled"""
        return {
            valve for valve in valves if valve.dp.stack and valve.dp.stack.root_name
        }

    def reset_peer_distances(self):
        """Recalculates the towards and away ports for this node"""
        self.towards_root_ports = set()
        self.chosen_towards_ports = set()
        self.chosen_towards_port = None

        self.away_ports = set()
        self.inactive_away_ports = set()
        self.pruned_away_ports = set()

        all_peer_ports = set(self.stack.canonical_up_ports())
        if self.stack.is_root():
            self.away_ports = all_peer_ports
        else:
            port_peer_distances = {
                port: len(port.stack["dp"].stack.shortest_path_to_root())
                for port in all_peer_ports
            }
            shortest_peer_distance = None
            for port, port_peer_distance in port_peer_distances.items():
                if shortest_peer_distance is None:
                    shortest_peer_distance = port_peer_distance
                    continue
                shortest_peer_distance = min(shortest_peer_distance, port_peer_distance)
            self.towards_root_ports = {
                port
                for port, port_peer_distance in port_peer_distances.items()
                if port_peer_distance == shortest_peer_distance
            }

            self.away_ports = all_peer_ports - self.towards_root_ports

            if self.towards_root_ports:
                # Generate a shortest path to calculate the chosen connection to root
                shortest_path = self.stack.shortest_path_to_root()
                # Choose the port that is connected to peer DP
                if shortest_path and len(shortest_path) > 1:
                    first_peer_dp = shortest_path[1]
                else:
                    first_peer_port = self.stack.canonical_port_order(
                        self.towards_root_ports
                    )[0]
                    first_peer_dp = first_peer_port.stack["dp"].name
                # The chosen towards ports are the ports through the chosen peer DP
                self.chosen_towards_ports = {
                    port
                    for port in self.towards_root_ports
                    if port.stack["dp"].name == first_peer_dp
                }  # pytype: disable=attribute-error

            if self.chosen_towards_ports:
                self.chosen_towards_port = self.stack.canonical_up_ports(
                    self.chosen_towards_ports
                )[0]

            # Away ports are all the remaining (non-towards) ports
            self.away_ports = all_peer_ports - self.towards_root_ports

        if self.away_ports:
            # Get inactive away ports, ports whose peers have a better path to root
            self.inactive_away_ports = {
                port
                for port in self.away_ports
                if not self.stack.is_in_path(
                    port.stack["dp"].name, self.stack.root_name
                )
            }

            # Get pruned away ports, redundant ports for each adjacent DP
            ports_by_dp = defaultdict(list)
            for port in self.away_ports:
                ports_by_dp[port.stack["dp"]].append(port)
            for ports in ports_by_dp.values():
                remote_away_ports = self.stack.canonical_up_ports(
                    [port.stack["port"] for port in ports]
                )
                self.pruned_away_ports.update(
                    [
                        port.stack["port"]
                        for port in remote_away_ports
                        if port != remote_away_ports[0]
                    ]
                )

        return self.chosen_towards_ports

    def update_stack_topo(self, event, dp, port):
        """
        Update the stack topo according to the event.

        Args:
            event (bool): True if the port is UP
            dp (DP): DP object
            port (Port): The port being brought UP/DOWN
        """
        self.stack.modify_link(dp, port, event)
        towards_ports = self.reset_peer_distances()
        if towards_ports:
            self.logger.info("shortest path to root is via %s" % towards_ports)
        else:
            self.logger.info("no path available to root")

    def default_port_towards(self, dp_name):
        """
        Default shortest path towards the provided destination, via direct shortest path

        Args:
            dp_name (str): Destination DP
        Returns:
           Port: port from current node that is shortest directly towards destination
        """
        return self.stack.shortest_path_port(dp_name)

    def relative_port_towards(self, dp_name):
        """
        Returns the shortest path towards provided destination, via either the root or away paths

        Args:
            dp_name (str): Destination DP
        Returns:
            Port: port from current node that is towards/away the destination DP depending on
                relative position of the current node
        """
        if not self.stack.shortest_path_to_root():
            # No known path from current node to root, use default
            return self.default_port_towards(dp_name)
        if self.stack.name == dp_name:
            # Current node is the destination node, use default
            return self.default_port_towards(dp_name)
        path_to_root = self.stack.shortest_path_to_root(dp_name)
        if path_to_root and self.stack.name in path_to_root:
            # Current node is a transit node between root & destination, direct path to destination
            away_dp = path_to_root[path_to_root.index(self.stack.name) - 1]
            for port in self.away_ports:
                if port.stack["dp"].name == away_dp and not self.is_pruned_port(port):
                    return port
            return None
        # Otherwise, head towards the root, path to destination via root
        return self.chosen_towards_port

    def edge_learn_port_towards(self, pkt_meta, edge_dp):
        """
        Returns the port towards the edge DP

        Args:
            pkt_meta (PacketMeta): Packet on the edge DP
            edge_dp (DP): Edge DP that received the packet
        Returns:
            Port: Port towards the edge DP via some stack chosen metric
        """
        if pkt_meta.vlan.edge_learn_stack_root:
            return self.relative_port_towards(edge_dp.name)
        return self.default_port_towards(edge_dp.name)

    def tunnel_outport(self, src_dp, dst_dp, dst_port):
        """
        Returns the output port for the current stack node for the tunnel path

        Args:
            src_dp (str): Source DP name of the tunnel
            dst_dp (str): Destination DP name of the tunnel
            dst_port (int): Destination port of the tunnel
        Returns:
            int: Output port number for the current node of the tunnel
        """
        if not self.stack.is_in_path(src_dp, dst_dp):
            # No known path from the source to destination DP, so no port to output
            return None
        out_port = self.default_port_towards(dst_dp)
        if self.stack.name == dst_dp:
            # Current stack node is the destination, so output to the tunnel destination port
            out_port = dst_port
        elif out_port:
            out_port = out_port.number
        return out_port

    def update_health(self, now, last_live_times, update_time):
        """
        Returns whether the current stack node is healthy, a healthy stack node
            is one that attempted connected recently, or was known to be running
            recently, has all LAGs UP and any stack port UP

        Args:
            now (float): Current time
            last_live_times (dict): Last live time value for each DP
            update_time (int): Stack root update interval time
        Returns:
            bool: True if current stack node is healthy
        """
        prev_health = self.stack.dyn_healthy_info
        new_health, reason = self.stack.update_health(now, last_live_times, update_time)
        if prev_health != self.stack.dyn_healthy_info:
            health = "HEALTHY" if new_health else "UNHEALTHY"
            self.logger.info(
                "Stack node %s %s (%s)" % (self.stack.name, health, reason)
            )
        return new_health

    @staticmethod
    def nominate_stack_root(
        root_valve, other_valves, now, last_live_times, update_time
    ):
        """
        Nominate a new stack root

        Args:
            root_valve (Valve): Previous/current root Valve object
            other_valves (list): List of other valves (not including previous root)
            now (float): Current time
            last_live_times (dict): Last live time value for each DP
            update_time (int): Stack root update interval time
        Returns:
            str: Name of the new elected stack root
        """
        stack_valves = {valve for valve in other_valves if valve.dp.stack}
        if root_valve:
            stack_valves = {root_valve}.union(stack_valves)

        # Create lists of healthy and unhealthy root candidates
        healthy_valves = []
        unhealthy_valves = []
        for valve in stack_valves:
            if valve.dp.stack.is_root_candidate():
                healthy = valve.stack_manager.update_health(
                    now, last_live_times, update_time
                )
                if healthy:
                    healthy_valves.append(valve)
                elif valve.dp.stack.dyn_healthy_info[0]:
                    unhealthy_valves.append(valve)

        if not healthy_valves and not unhealthy_valves:
            # No root candidates/stack valves, so no nomination
            return None

        # Choose a candidate valve to be the root
        if healthy_valves:
            # Healthy valves exist, so pick a healthy valve as root
            new_root_name = None
            if root_valve:
                new_root_name = root_valve.dp.name
            if root_valve not in healthy_valves:
                # Need to pick a new healthy root if current root not healthy
                stacks = [valve.dp.stack for valve in healthy_valves]
                _, new_root_name = stacks[0].nominate_stack_root(stacks)
        else:
            # No healthy stack roots, so forced to choose a bad root
            new_root_name = None
            if root_valve:
                # Current root is unhealthy along with all other roots, so keep root the same
                new_root_name = root_valve.dp.name
            if root_valve not in unhealthy_valves:
                # Pick the best unhealthy root
                stacks = [valve.dp.stack for valve in unhealthy_valves]
                _, new_root_name = stacks[0].nominate_stack_root(stacks)

        return new_root_name

    def consistent_roots(self, expected_root_name, valve, other_valves):
        """Returns true if all the stack nodes have the root configured correctly"""
        stacked_valves = {valve}.union(self.stacked_valves(other_valves))
        for stack_valve in stacked_valves:
            if stack_valve.dp.stack.root_name != expected_root_name:
                return False
        return True

    def stack_ports(self):
        """Yield the stack ports of this stack node"""
        for port in self.stack.ports:
            yield port

    @staticmethod
    def is_stack_port(port):
        """Return whether the port is a stack port"""
        return bool(port.stack)

    def is_away(self, port):
        """Return whether the port is an away port for the node"""
        return port in self.away_ports

    def is_towards_root(self, port):
        """Return whether the port is a port towards the root for the node"""
        return port in self.towards_root_ports

    def is_selected_towards_root_port(self, port):
        """Return true if the port is the chosen towards root port"""
        return port == self.chosen_towards_port

    def is_pruned_port(self, port):
        """Return true if the port is to be pruned"""
        if self.is_towards_root(port):
            return not self.is_selected_towards_root_port(port)
        if self.is_away(port):
            if self.pruned_away_ports:
                return port in self.pruned_away_ports
            return False
        return True

    def adjacent_stack_ports(self, peer_dp):
        """Return list of ports that connect to an adjacent DP"""
        return [port for port in self.stack.ports if port.stack["dp"] == peer_dp]

    def acl_update_tunnel(self, acl):
        """Return ofmsgs for all tunnels in an ACL with a tunnel rule"""
        ofmsgs = []
        source_vids = defaultdict(list)
        for _id, tunnel_dest in acl.tunnel_dests.items():
            dst_dp, dst_port = tunnel_dest["dst_dp"], tunnel_dest["dst_port"]
            # Update the tunnel rules for each tunnel action specified
            updated_sources = []
            updated_reverse_sources = []
            for source_id, source in acl.tunnel_sources.items():
                # We loop through each tunnel source in a single ACL instance and update the info
                src_dp, src_port = source["dp"], source["port"]
                in_port = self.tunnel_outport(dst_dp, src_dp, src_port)
                out_port = self.tunnel_outport(src_dp, dst_dp, dst_port)
                updated = False
                if out_port is None and dst_port is None and dst_dp == self.dp.name:
                    # Will need to update at most once, to ensure the correct rules
                    # get populated in the destination DP for a tunnel that outputs
                    # to just a DP
                    updated = acl.update_source_tunnel_rules(
                        self.stack.name, source_id, _id, out_port, self.output_table
                    )
                elif out_port:
                    updated = acl.update_source_tunnel_rules(
                        self.stack.name, source_id, _id, out_port, self.output_table
                    )
                if updated:
                    if self.stack.name == src_dp:
                        # We need to re-build and apply the whole ACL
                        source_vids[source_id].append(_id)
                    else:
                        # We only need to re-build and apply the tunnel
                        updated_sources.append(source_id)
                reverse_updated = False
                if src_port is None and in_port is None and src_dp == self.dp.name:
                    reverse_updated = acl.update_reverse_tunnel_rules(
                        self.stack.name, source_id, _id, in_port, self.output_table
                    )
                elif in_port:
                    reverse_updated = acl.update_reverse_tunnel_rules(
                        self.stack.name, source_id, _id, in_port, self.output_table
                    )
                if reverse_updated:
                    if acl.requires_reverse_tunnel(_id):
                        # Update the reverse tunnel rules if the tunnel is configured to have them
                        updated_reverse_sources.append(source_id)
            # The tunnel in the ACL does not have a source on this stack instance, so
            #   we only need to re-build the special tunnel forwarding rule.
            for source_id in updated_sources:
                ofmsgs.extend(
                    self.acl_manager.build_tunnel_rules_ofmsgs(source_id, _id, acl)
                )
            for source_id in updated_reverse_sources:
                ofmsgs.extend(
                    self.acl_manager.build_reverse_tunnel_rules_ofmsgs(
                        source_id, _id, acl
                    )
                )
        # If a tunnel is updated, but the source is configured as the current DP
        #   then we will also need to re-build the rest of the ACL rules aswell.
        for source_id, vids in source_vids.items():
            for vid in vids:
                ofmsgs.extend(
                    self.acl_manager.build_tunnel_acl_rule_ofmsgs(source_id, vid, acl)
                )
        return ofmsgs

    def add_tunnel_acls(self):
        """Returns ofmsgs installing the tunnel path rules"""
        ofmsgs = []
        if self.tunnel_acls:
            for acl in self.tunnel_acls:
                ofmsgs.extend(self.acl_update_tunnel(acl))
        return ofmsgs

    def add_port(self, port):
        """Need to add tunnel if port comes up with tunnel ACLs."""
        if not port.stack and port.tunnel_acls():
            return self.add_tunnel_acls()
        return []



"""A docker-based mininet host"""

import json
import operator
import os
import pty
import select
from subprocess import PIPE, STDOUT
from functools import reduce

# pylint: disable=import-error
from mininet.log import error, debug
from mininet.node import Host
from mininet.util import quietRun, errRun

from clib.mininet_test_util import DEVNULL

DEFAULT_NETWORK = "none"
DEFAULT_PREFIX = "mininet"
STARTUP_TIMEOUT_MS = 20000

# pylint: disable=too-many-instance-attributes


class DockerHost(Host):
    """Mininet host that encapsulates execution in a docker container"""

    master = None
    shell = None
    slave = None
    name = None
    inNamespace = None
    pollOut = None
    stdout = None
    execed = None
    lastCmd = None  # pylint: disable=invalid-name
    readbuf = None
    lastPid = None
    pid = None
    waiting = None
    stdin = None
    active_pipe = None
    active_pipe_returncode = None
    image = None
    tmpdir = None
    env_vars = None
    vol_maps = None
    prefix = None
    startup_timeout_ms = None
    container = None
    pollIn = None
    active_log = None
    ps1 = chr(127)

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        name,
        image=None,
        tmpdir=None,
        prefix=None,
        env_vars=None,
        vol_maps=None,
        startup_timeout_ms=STARTUP_TIMEOUT_MS,
        network=None,
        **kwargs
    ):
        self.image = image
        self.tmpdir = tmpdir
        self.prefix = prefix
        if env_vars is None:
            env_vars = []
        self.env_vars = env_vars
        if vol_maps is None:
            vol_maps = []
        self.vol_maps = vol_maps
        self.network = network
        self.startup_timeout_ms = startup_timeout_ms
        self.name = name
        self.pullImage()
        Host.__init__(self, name, **kwargs)

    def pullImage(self):  # pylint: disable=invalid-name
        "Pull docker image if necessary"
        if self.image not in quietRun("docker images"):
            error(
                "%s: docker image" % self.name,
                self.image,
                "not available locally - pulling\n",
            )
            _out, err, code = errRun("docker", "pull", self.image)
            if err or code:
                error("docker pull failed with error", code, err, "\n")

    # pylint: disable=invalid-name
    def startShell(self, mnopts=None):
        """Start a shell process for running commands."""
        if self.shell:
            error("shell is already running")
            return

        assert mnopts is None, "mnopts not supported for DockerHost"

        self.container = "%s-%s" % (self.prefix, self.name)

        debug('Starting container %s with image "%s".' % (self.container, self.image))

        self.kill(purge=True)

        container_tmp_dir = os.path.join(os.path.abspath(self.tmpdir), "tmp")
        tmp_volume = container_tmp_dir + ":/tmp"

        base_cmd = [
            "docker",
            "run",
            "-ti",
            "--privileged",
            "--entrypoint",
            "env",
            "-h",
            self.name,
            "--name",
            self.container,
        ]
        opt_args = ["--net=%s" % self.network]
        env_vars = self.env_vars + ["TERM=dumb", "PS1=%s" % self.ps1]
        env_args = reduce(operator.add, (["--env", var] for var in env_vars), [])
        vol_args = reduce(
            operator.add, (["-v", var] for var in self.vol_maps), ["-v", tmp_volume]
        )
        image_args = [self.image, "bash", "--norc", "-is", "mininet:" + self.name]
        cmd = base_cmd + opt_args + env_args + vol_args + image_args
        self.master, self.slave = pty.openpty()
        debug(
            'docker command "%s", fd %d, fd %d'
            % (" ".join(cmd), self.master, self.slave)
        )
        try:
            self.shell = self._popen(
                cmd, stdin=self.slave, stdout=self.slave, stderr=self.slave
            )
            self.stdin = os.fdopen(self.master, "r")
            self.stdout = self.stdin
            self.pollOut = select.poll()  # pylint: disable=invalid-name
            self.pollOut.register(self.stdout)  # pylint: disable=no-member
            self.outToNode[self.stdout.fileno()] = self  # pylint: disable=no-member
            self.pollIn = select.poll()  # pylint: disable=invalid-name
            self.pollIn.register(
                self.stdout, select.POLLIN
            )  # pylint: disable=no-member
            self.inToNode[self.stdin.fileno()] = self  # pylint: disable=no-member
            self.execed = False
            self.lastCmd = None  # pylint: disable=invalid-name
            self.lastPid = None  # pylint: disable=invalid-name
            self.readbuf = ""
            self.waiting = True
            data = ""
            while True:
                data = self.read(1)
                if data[-1] == self.ps1:
                    break
            self.readbuf = ""
            self.waiting = False
        except Exception:
            error("docker cmd: %s" % " ".join(cmd))
            if self.shell.returncode:
                error("returncode: %d" % self.shell.returncode)
            if self.shell:
                self.shell.poll()
            raise

        self.pid = self.inspect_pid()
        debug(
            "Container %s created pid %s/%s."
            % (self.container, self.pid, self.shell.pid)
        )

        self.cmd("unset HISTFILE; stty -echo; set +m")  # pylint: disable=no-member

    def kill(self, purge=False):
        """Kill a container."""
        debug("killing container %s." % self.container)
        if purge:
            kill_cmd = ["docker", "rm", "-f", self.container]
        else:
            kill_cmd = ["docker", "kill", self.container]
        kill_pipe = None
        try:
            kill_pipe = self._popen(kill_cmd, stdin=DEVNULL, stdout=PIPE, stderr=STDOUT)
            kill_pipe.stdout.readlines()
            kill_pipe.stdout.close()
        except Exception:
            if kill_pipe:
                kill_pipe.poll()
            raise

    def inspect_pid(self):
        """Return container PID."""
        pid_pipe = None
        try:
            pid_cmd = ["docker", "inspect", "--format={{ .State.Pid }}", self.container]
            pid_pipe = self._popen(pid_cmd, stdin=DEVNULL, stdout=PIPE, stderr=STDOUT)
            ps_out = pid_pipe.stdout.readlines()
            pid_pipe.stdout.close()
            return int(ps_out[0])
        except Exception:
            if pid_pipe is not None:
                pid_pipe.poll()
            raise

    def open_log(self):
        """Open a log file for writing and return it."""
        # pylint: disable=consider-using-with
        return open(os.path.join(self.tmpdir, "activate.log"), "w")

    def activate(self):
        """Activate a container and return STDOUT to it."""
        assert not self.active_pipe, "container %s already activated" % self.container
        debug("activating container %s." % self.container)
        inspect_cmd = ["docker", "inspect", "--format={{json .Config}}", self.image]
        inspect_pipe = None
        try:
            inspect_pipe = self._popen(
                inspect_cmd, stdin=DEVNULL, stdout=PIPE, stderr=STDOUT
            )
            config_json = inspect_pipe.stdout.readlines()
            inspect_pipe.stdout.close()
            assert len(config_json) == 1, "Expected 1 config line, found %s" % len(
                config_json
            )
            config = json.loads(config_json[0].decode())
            entryconfig = config.get("Entrypoint")
            entrypoint = entryconfig if entryconfig else ["/usr/bin/env"]
            cmd = config["Cmd"] if "Cmd" in config else []
            docker_cmd = entrypoint + (cmd if cmd else [])
            debug("logging to activate.log for %s" % docker_cmd)
            stdout = self.open_log()
            self.active_log = stdout
        except Exception:
            if inspect_pipe:
                inspect_pipe.poll()
            raise
        self.active_pipe_returncode = None
        self.active_pipe = self.popen(
            docker_cmd, stdin=DEVNULL, stdout=stdout, stderr=STDOUT
        )
        pipe_out = self.active_pipe.stdout
        out_fd = pipe_out.fileno() if pipe_out else None
        debug(
            "Active_pipe container %s pid %s fd %s"
            % (self.container, self.active_pipe.pid, out_fd)
        )
        return self.active_pipe

    def wait(self):
        """Wait for an activated container to terminate."""
        try:
            if self.active_pipe_returncode is not None:
                return self.active_pipe_returncode
            debug("Waiting for container %s." % self.container)
            assert self.active_pipe, "container not activated"
            self.active_pipe.communicate()
            self.active_pipe.returncode = self.active_pipe.wait()
            self.terminate()
            return self.active_pipe_returncode
        except Exception as err:
            error("Exception waiting for %s: %s" % (self.container, err))
            self.terminate()
            raise

    def read(self, size=1024):
        """Read from an activated container."""
        poll_results = self.pollIn.poll(self.startup_timeout_ms)
        data_ready = poll_results and (poll_results[0][1] & select.POLLIN)
        assert data_ready, "Timeout waiting for read data on %d after %ds" % (
            self.stdout.fileno(),
            self.startup_timeout_ms / 1e3,
        )
        return Host.read(self, size)

    def terminate(self):
        """Override Mininet terminate() to partially avoid pty leak."""
        debug(
            "Terminating container %s, shell %s, pipe %s"
            % (self.container, self.shell, self.active_pipe)
        )
        if self.slave:
            os.close(self.slave)
            self.slave = None
        if self.shell is not None:
            self.stdin.close()
            self.stdin = None
            self.master = None
            if self.shell.returncode is None:
                self.shell.kill()
                self.shell.poll()
            self.kill()
            self.shell = None
        if self.active_pipe:
            if self.active_pipe.stdout:
                self.active_pipe.stdout.close()
            if self.active_pipe.returncode is None:
                self.active_pipe.kill()
                self.active_pipe.poll()
            self.active_pipe_returncode = self.active_pipe.returncode
            self.active_pipe = None
            if self.active_log:
                self.active_log.close()
                self.active_log = None
        self.cleanup()  # pylint: disable=no-member
        return self.active_pipe_returncode

    def popen(self, *args, **kwargs):
        """Return a Popen() object in node's namespace
        args: Popen() args, single list, or string
        kwargs: Popen() keyword args"""
        # -t is necessary to prevent docker from buffering output. It might cause
        # problems with some commands like shells that then assume they can output
        # all sorts of crazy control characters b/c it's a terminal.
        mncmd = ["docker", "exec", "--env", "TERM=dumb", "-t", self.container]
        pipe = Host.popen(self, mncmd=mncmd, *args, **kwargs)
        if pipe:
            debug("docker pid %d: %s %s %s" % (pipe.pid, mncmd, args, kwargs))
        return pipe

    def _popen(self, cmd, **params):
        # Docker is different than mininet in that it doesn't handle signals like
        # a normal interactive terminal would. So, put it in a separate process group
        # so it doesn't receive stray SIGINTs, rather relying on the message sent
        # from the owning process through the pty.
        if "preexec_fn" not in params:
            params["preexec_fn"] = os.setpgrp
        pipe = super()._popen(cmd, **params)
        if pipe:
            stdout = pipe.stdout
            out_fd = pipe.stdout.fileno() if stdout else None
            debug("docker pid %d: %s, fd %s" % (pipe.pid, cmd, out_fd))
        return pipe


def make_docker_host(
    image, prefix=DEFAULT_PREFIX, network=DEFAULT_NETWORK, startup_timeout_ms=None
):
    """Utility function to create a docker-host class that can be passed to mininet"""

    class _ImageHost(DockerHost):
        """Internal class that represents a docker image host"""

        def __init__(self, *args, **kwargs):
            host_name = args[0]
            kwargs["image"] = image
            assert kwargs["tmpdir"], "tmpdir required for docker host"
            kwargs["tmpdir"] = os.path.join(kwargs["tmpdir"], host_name)
            kwargs["prefix"] = prefix
            kwargs["network"] = network
            if startup_timeout_ms:
                kwargs["startup_timeout_ms"] = startup_timeout_ms
            elif "DOCKER_STARTUP_TIMEOUT_MS" in os.environ:
                env_val = os.environ["DOCKER_STARTUP_TIMEOUT_MS"]
                if env_val:
                    kwargs["startup_timeout_ms"] = int(env_val)
            super().__init__(*args, **kwargs)

    return _ImageHost



"""Configuration for ACLs."""

# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import netaddr
from os_ken.ofproto import ether

from faucet import valve_of
from faucet import valve_acl
from faucet.valve_of import MATCH_FIELDS, OLD_MATCH_FIELDS
from faucet.conf import Conf, test_config_condition, InvalidConfigError
from faucet.valve_table import wildcard_table


class ACL(Conf):
    """Contains the state for an ACL, including the configuration.

ACL Config

ACLs are configured under the 'acls' configuration block. The acls block
contains a dictionary of individual acls each keyed by its name.

Each acl contains a list of rules, a packet will have the first matching rule
applied to it.

Each rule is a dictionary containing the single key 'rule' with the value the
matches and actions for the rule.

The matches are key/values based on the ryu RESTFul API.
The key 'actions' contains a dictionary with keys/values as follows:

 * allow (int): if 1 allow the packet to continue through the Faucet \
       pipeline, if 0 drop the packet.
 * force_port_vlan (int): if 1, do not verify the VLAN/port association \
       for this packet and override any VLAN ACL on the forced VLAN.
 * meter (str): meter to apply to the packet
 * output (dict): used to output a packet directly. details below.
 * cookie (int): set flow cookie to this value on this flow

The output action contains a dictionary with the following elements:

 * tunnel (dict): the tunnel formation, creates a tunnel from the applied port(s) \
       to the specified destination
 * port (int or string): the port to output the packet to
 * ports (list): a list of the ports (int or string) to output the packet to
 * set_fields (list): a list of fields to set with values
 * pop_vlans: (int): pop the packet vlan before outputting
 * vlan_vid: (int): push the vlan vid on the packet when outputting
 * vlan_vids: (list): push the list of vlans on the packet when outputting, with option eth_type
 * swap_vid (int): rewrite the vlan vid of the packet when outputting
 * failover (dict): Output with a failover port (experimental)
"""

    defaults = {
        "rules": None,
        "exact_match": False,
        "dot1x_assigned": False,
    }
    defaults_types = {
        "rules": list,
        "exact_match": bool,
        "dot1x_assigned": bool,
    }
    rule_types = {
        "cookie": int,
        "actions": dict,
        "description": str,
    }
    actions_types = {
        "meter": str,
        "mirror": (str, int),
        "output": (dict, list),
        "allow": int,
        "force_port_vlan": int,
        "ct": dict,
    }
    output_actions_types = {
        "tunnel": dict,
        "port": (str, int),
        "ports": list,
        "failover": dict,
        "set_fields": list,
        "pop_vlans": int,
        "swap_vid": int,
        "vlan_vid": int,
        "vlan_vids": list,
    }
    ct_action_types = {
        "flags": int,
        "alg": int,
        "table": int,
        "zone": int,
        "zone_src": int,
        "clear": bool,
        "nat": dict,
        "mark": int,
        "label": int,
    }
    ct_action_nat_types = {
        "flags": int,
        "range_ipv4_min": str,
        "range_ipv4_max": str,
        "range_ipv6_min": str,
        "range_ipv6_max": str,
        "range_proto_min": int,
        "range_proto_max": int,
    }
    tunnel_types = {
        "type": (str, None),
        "tunnel_id": (str, int, None),
        "dp": str,
        "port": (str, int, None),
        "exit_instructions": (list, None),
        "maintain_encapsulation": bool,
        "bi_directional": bool,
        "reverse": bool,
    }

    mutable_attrs = frozenset(["tunnel_sources"])

    def __init__(self, _id, dp_id, conf):
        self.rules = []
        self.exact_match = None
        self.dot1x_assigned = None
        self.meter = False
        self.matches = {}
        self.set_fields = set()
        self._ports_resolved = False

        # Tunnel info maintains the tunnel output information for each tunnel rule
        self.tunnel_dests = {}
        # Tunnel sources is a list of the sources in the network for this ACL
        self.tunnel_sources = {}
        # Tunnel rules is the rules for each tunnel in the ACL for each source
        self.dyn_tunnel_rules = {}
        self.dyn_reverse_tunnel_rules = {}

        for match_fields in (MATCH_FIELDS, OLD_MATCH_FIELDS):
            self.rule_types.update({match: (str, int) for match in match_fields})
        conf = copy.deepcopy(conf)
        if isinstance(conf, dict):
            rules = conf.get("rules", [])
        elif isinstance(conf, list):
            rules = conf
            conf = {}
        else:
            raise InvalidConfigError("ACL conf is an invalid type %s" % _id)
        conf["rules"] = []
        for rule in rules:
            normalized_rule = rule
            if isinstance(rule, dict):
                normalized_rule = rule.get("rule", rule)
                if normalized_rule is None:
                    normalized_rule = {k: v for k, v in rule.items() if v is not None}
            test_config_condition(
                not isinstance(normalized_rule, dict),
                ("ACL rule is %s not %s (%s)" % (type(normalized_rule), dict, rules)),
            )
            conf["rules"].append(normalized_rule)
        super().__init__(_id, dp_id, conf)

    def finalize(self):
        self._ports_resolved = True
        super().finalize()

    def check_config(self):
        test_config_condition(not self.rules, "no rules found for ACL %s" % self._id)
        for rule in self.rules:
            self._check_conf_types(rule, self.rule_types)
            for rule_field, rule_conf in rule.items():
                if rule_field == "cookie":
                    test_config_condition(
                        rule_conf < 0 or rule_conf > 2**16,
                        "rule cookie value must be 0-2**16",
                    )
                elif rule_field == "actions":
                    test_config_condition(
                        not rule_conf, "Missing rule actions in ACL %s" % self._id
                    )
                    self._check_conf_types(rule_conf, self.actions_types)
                    for action_name, action_conf in rule_conf.items():
                        if action_name == "output":
                            if isinstance(action_conf, (list, tuple)):
                                # New ordered format
                                for subconf in action_conf:
                                    # Make sure only one specified action per list element
                                    test_config_condition(
                                        len(subconf) > 1,
                                        "ACL ordered output must have only one action per element",
                                    )
                                    # Ensure correct action format
                                    self._check_conf_types(
                                        subconf, self.output_actions_types
                                    )
                            else:
                                # Old format
                                self._check_conf_types(
                                    action_conf, self.output_actions_types
                                )
                        elif action_name == "ct":
                            self._check_conf_types(action_conf, self.ct_action_types)
                            # if clear set, make sure nothing else is
                            if "clear" in action_conf and action_conf["clear"]:
                                test_config_condition(
                                    len(action_conf) != 1,
                                    "no other parameters can be set when 'clear' set on "
                                    "conntrack ACL",
                                )
                            else:
                                test_config_condition(
                                    "table" not in action_conf,
                                    "required parameter 'table' not set for conntrack ACL",
                                )
                                test_config_condition(
                                    "zone" not in action_conf,
                                    "required parameter 'zone' not set for conntrack ACL",
                                )
                            if "nat" in action_conf:
                                self._check_conf_types(
                                    action_conf["nat"], self.ct_action_nat_types
                                )

    def build(self, meters, vid, port_num):
        """Check that ACL can be built from config."""

        self.matches = {}
        self.set_fields = set()
        self.meter = False
        if self.rules:
            try:
                ofmsgs = valve_acl.build_acl_ofmsgs(
                    [self],
                    wildcard_table,
                    [valve_of.goto_table(wildcard_table)],
                    [valve_of.goto_table(wildcard_table)],
                    2**16 - 1,
                    meters,
                    self.exact_match,
                    vlan_vid=vid,
                    port_num=port_num,
                )
            except (netaddr.core.AddrFormatError, KeyError, ValueError) as err:
                raise InvalidConfigError from err
            test_config_condition(not ofmsgs, "OF messages is empty")
            for ofmsg in ofmsgs:
                try:
                    valve_of.verify_flowmod(ofmsg)
                except (KeyError, ValueError) as err:
                    raise InvalidConfigError from err
                except Exception as err:
                    raise err
                if valve_of.is_flowmod(ofmsg):
                    apply_actions = []
                    for inst in ofmsg.instructions:
                        if valve_of.is_apply_actions(inst):
                            apply_actions.extend(inst.actions)
                        elif valve_of.is_meter(inst):
                            self.meter = True
                    for action in apply_actions:
                        if valve_of.is_set_field(action):
                            self.set_fields.add(action.key)
                    for match, value in ofmsg.match.items():
                        has_mask = isinstance(value, (tuple, list))
                        if has_mask or match not in self.matches:
                            self.matches[match] = has_mask
        for tunnel_rules in self.tunnel_dests.values():
            if "exit_instructions" in tunnel_rules:
                exit_inst = tunnel_rules["exit_instructions"]
                try:
                    ofmsgs = valve_acl.build_tunnel_ofmsgs(exit_inst, wildcard_table, 1)
                except (netaddr.core.AddrFormatError, KeyError, ValueError) as err:
                    raise InvalidConfigError from err
                test_config_condition(not ofmsgs, "OF messages is empty")
                for ofmsg in ofmsgs:
                    try:
                        valve_of.verify_flowmod(ofmsg)
                    except (KeyError, ValueError) as err:
                        raise InvalidConfigError from err
                    except Exception as err:
                        raise err
                    if valve_of.is_flowmod(ofmsg):
                        apply_actions = []
                        for inst in ofmsg.instructions:
                            if valve_of.is_apply_actions(inst):
                                apply_actions.extend(inst.actions)
                            elif valve_of.is_meter(inst):
                                self.meter = True
                        for action in apply_actions:
                            if valve_of.is_set_field(action):
                                self.set_fields.add(action.key)
                        for match, value in ofmsg.match.items():
                            has_mask = isinstance(value, (tuple, list))
                            if has_mask or match not in self.matches:
                                self.matches[match] = has_mask
        return (self.matches, self.set_fields, self.meter)

    def get_meters(self):
        """Yield meters for each rule in ACL"""
        for rule in self.rules:
            if "actions" not in rule or "meter" not in rule["actions"]:
                continue
            yield rule["actions"]["meter"]

    def get_mirror_destinations(self):
        """Yield mirror destinations for each rule in ACL"""
        for rule in self.rules:
            if "actions" not in rule or "mirror" not in rule["actions"]:
                continue
            yield rule["actions"]["mirror"]

    def _resolve_ordered_output_ports(
        self, output_list, resolve_port_cb, resolve_tunnel_objects
    ):
        """Resolve output actions in the ordered list format"""
        result = []
        for action in output_list:
            for key, value in action.items():
                if key == "tunnel":
                    tunnel = value
                    # Fetch tunnel items from the tunnel output dict
                    test_config_condition(
                        "dp" not in tunnel, "ACL (%s) tunnel DP not defined" % self._id
                    )
                    tunnel_dp = tunnel["dp"]
                    tunnel_port = tunnel.get("port", None)
                    tunnel_id = tunnel.get("tunnel_id", None)
                    tunnel_type = tunnel.get("type", "vlan")
                    tunnel_exit_instructions = tunnel.get("exit_instructions", [])
                    tunnel_direction = tunnel.get("bi_directional", False)
                    tunnel_maintain = tunnel.get("maintain_encapsulation", False)
                    tunnel_reverse = tunnel.get("reverse", False)
                    test_config_condition(
                        tunnel_reverse and tunnel_direction,
                        (
                            "Tunnel ACL %s cannot contain values for the fields"
                            "`bi_directional` and `reverse` at the same time" % self._id
                        ),
                    )
                    # Resolve the tunnel items
                    dst_dp, dst_port, tunnel_id = resolve_tunnel_objects(
                        tunnel_dp, tunnel_port, tunnel_id
                    )
                    # Compile the tunnel into an easy-access dictionary
                    tunnel_dict = {
                        "dst_dp": dst_dp,
                        "dst_port": dst_port,
                        "tunnel_id": tunnel_id,
                        "type": tunnel_type,
                        "exit_instructions": tunnel_exit_instructions,
                        "bi_directional": tunnel_direction,
                        "maintain_encapsulation": tunnel_maintain,
                        "reverse": tunnel_reverse,
                    }
                    self.tunnel_dests[tunnel_id] = tunnel_dict
                    result.append({key: tunnel_id})
                elif key == "port":
                    port_name = value
                    port = resolve_port_cb(port_name)
                    test_config_condition(
                        not port,
                        "ACL (%s) output port undefined in DP: %s"
                        % (self._id, self.dp_id),
                    )
                    result.append({key: port})
                elif key == "ports":
                    resolved_ports = [resolve_port_cb(p) for p in value]
                    test_config_condition(
                        None in resolved_ports,
                        "ACL (%s) output port(s) not defined in DP: %s"
                        % (self._id, self.dp_id),
                    )
                    result.append({key: resolved_ports})
                elif key == "failover":
                    failover = value
                    test_config_condition(
                        not isinstance(failover, dict), ("failover is not a dictionary")
                    )
                    failover_dict = {}
                    for failover_name, failover_values in failover.items():
                        if failover_name == "ports":
                            resolved_ports = [
                                resolve_port_cb(p) for p in failover_values
                            ]
                            test_config_condition(
                                None in resolved_ports,
                                "ACL (%s) failover port(s) not defined in DP: %s"
                                % (self._id, self.dp_id),
                            )
                            failover_dict[failover_name] = resolved_ports
                        else:
                            failover_dict[failover_name] = failover_values
                    result.append({key: failover_dict})
                else:
                    result.append(action)
        return result

    def _resolve_output_ports(
        self, action_conf, resolve_port_cb, resolve_tunnel_objects
    ):
        """Resolve the values for output actions in the ACL"""
        if isinstance(action_conf, (list, tuple)):
            return self._resolve_ordered_output_ports(
                action_conf, resolve_port_cb, resolve_tunnel_objects
            )
        result = {}
        test_config_condition(
            "vlan_vid" in action_conf and "vlan_vids" in action_conf,
            "ACL %s has both vlan_vid and vlan_vids defined" % self._id,
        )
        test_config_condition(
            "port" in action_conf and "ports" in action_conf,
            "ACL %s has both port and ports defined" % self._id,
        )
        for output_action, output_action_values in action_conf.items():
            if output_action == "tunnel":
                tunnel = output_action_values
                # Fetch tunnel items from the tunnel output dict
                test_config_condition(
                    "dp" not in tunnel, "ACL (%s) tunnel DP not defined" % self._id
                )
                tunnel_dp = tunnel["dp"]
                tunnel_port = tunnel.get("port", None)
                tunnel_id = tunnel.get("tunnel_id", None)
                tunnel_type = tunnel.get("type", "vlan")
                tunnel_exit_instructions = tunnel.get("exit_instructions", [])
                tunnel_direction = tunnel.get("bi_directional", False)
                tunnel_maintain = tunnel.get("maintain_encapsulation", False)
                tunnel_reverse = tunnel.get("reverse", False)
                test_config_condition(
                    tunnel_reverse and tunnel_direction,
                    (
                        "Tunnel ACL %s cannot contain values for the fields"
                        "`bi_directional` and `reverse` at the same time" % self._id
                    ),
                )
                # Resolve the tunnel items
                dst_dp, dst_port, tunnel_id = resolve_tunnel_objects(
                    tunnel_dp, tunnel_port, tunnel_id
                )
                # Compile the tunnel into an easy-access dictionary
                tunnel_dict = {
                    "dst_dp": dst_dp,
                    "dst_port": dst_port,
                    "tunnel_id": tunnel_id,
                    "type": tunnel_type,
                    "exit_instructions": tunnel_exit_instructions,
                    "bi_directional": tunnel_direction,
                    "maintain_encapsulation": tunnel_maintain,
                    "reverse": tunnel_reverse,
                }
                self.tunnel_dests[tunnel_id] = tunnel_dict
                result[output_action] = tunnel_id
            elif output_action == "port":
                port_name = output_action_values
                port = resolve_port_cb(port_name)
                test_config_condition(
                    not port,
                    (
                        "ACL (%s) output port undefined in DP: %s"
                        % (self._id, self.dp_id)
                    ),
                )
                result[output_action] = port
            elif output_action == "ports":
                resolved_ports = [resolve_port_cb(p) for p in output_action_values]
                test_config_condition(
                    None in resolved_ports,
                    (
                        "ACL (%s) output port(s) not defined in DP: %s"
                        % (self._id, self.dp_id)
                    ),
                )
                result[output_action] = resolved_ports
            elif output_action == "failover":
                failover = output_action_values
                test_config_condition(
                    not isinstance(failover, dict), ("failover is not a dictionary")
                )
                result[output_action] = {}
                for failover_name, failover_values in failover.items():
                    if failover_name == "ports":
                        resolved_ports = [resolve_port_cb(p) for p in failover_values]
                        test_config_condition(
                            None in resolved_ports,
                            (
                                "ACL (%s) failover port(s) not defined in DP: %s"
                                % (self._id, self.dp_id)
                            ),
                        )
                        result[output_action][failover_name] = resolved_ports
                    else:
                        result[output_action][failover_name] = failover_values
            else:
                result[output_action] = output_action_values
        return result

    def resolve_ports(self, resolve_port_cb, resolve_tunnel_objects):
        """Resolve the values for the actions of an ACL"""
        if self._ports_resolved:
            return
        for rule_conf in self.rules:
            if "actions" in rule_conf:
                actions_conf = rule_conf["actions"]
                resolved_actions = {}
                test_config_condition(
                    not isinstance(actions_conf, dict),
                    ("actions value is not a dictionary"),
                )
                for action_name, action_conf in actions_conf.items():
                    if action_name == "mirror":
                        resolved_port = resolve_port_cb(action_conf)
                        test_config_condition(
                            resolved_port is None,
                            (
                                "ACL (%s) mirror port is not defined in DP: %s"
                                % (self._id, self.dp_id)
                            ),
                        )
                        resolved_actions[action_name] = resolved_port
                    elif action_name == "output":
                        resolved_action = self._resolve_output_ports(
                            action_conf, resolve_port_cb, resolve_tunnel_objects
                        )
                        resolved_actions[action_name] = resolved_action
                    else:
                        resolved_actions[action_name] = action_conf
                rule_conf["actions"] = resolved_actions
        self._ports_resolved = True

    def requires_reverse_tunnel(self, tunnel_id):
        """Returns true if the tunnel requires a reverse pathway"""
        return self.tunnel_dests[tunnel_id]["bi_directional"]

    def get_num_tunnels(self):
        """Returns the number of tunnels specified in the ACL"""
        num_tunnels = 0
        for rule_conf in self.rules:
            if self.does_rule_contain_tunnel(rule_conf):
                output_conf = rule_conf["actions"]["output"]
                if isinstance(output_conf, list):
                    for action in output_conf:
                        for key in action:
                            if key == "tunnel":
                                num_tunnels += 1
                else:
                    if "tunnel" in output_conf:
                        num_tunnels += 1
        return num_tunnels

    def get_tunnel_rules(self, tunnel_id):
        """Return the list of rules that apply a specific tunnel ID"""
        rules = []
        for rule_conf in self.rules:
            if self.does_rule_contain_tunnel(rule_conf):
                output_conf = rule_conf["actions"]["output"]
                if isinstance(output_conf, (list, tuple)):
                    for action in output_conf:
                        for key, value in action.items():
                            if key == "tunnel" and value == tunnel_id:
                                rules.append(rule_conf)
                                continue
                else:
                    if output_conf["tunnel"] == tunnel_id:
                        rules.append(rule_conf)
        return rules

    @staticmethod
    def does_rule_contain_tunnel(rule_conf):
        """Return true if the ACL rule contains a tunnel"""
        if "actions" in rule_conf:
            if "output" in rule_conf["actions"]:
                output_conf = rule_conf["actions"]["output"]
                if isinstance(output_conf, (list, tuple)):
                    for action in output_conf:
                        for key in action:
                            if key == "tunnel":
                                return True
                else:
                    if "tunnel" in output_conf:
                        return True
        return False

    def is_tunnel_acl(self):
        """Return true if the ACL contains a tunnel"""
        if self.tunnel_dests:
            return True
        for rule_conf in self.rules:
            if self.does_rule_contain_tunnel(rule_conf):
                return True
        return False

    @staticmethod
    def _tunnel_source_id(source):
        """Return ID for a tunnel source."""
        return tuple(sorted(source.items()))

    def add_tunnel_source(self, dp_name, port, reverse=False, bi_directional=False):
        """Add a source dp/port pair for the tunnel ACL"""
        source = {
            "dp": dp_name,
            "port": port,
            "reverse": reverse,
            "bi_directional": bi_directional,
        }
        source_id = self._tunnel_source_id(source)
        self.tunnel_sources[source_id] = source
        for _id in self.tunnel_dests:
            self.dyn_tunnel_rules.setdefault(_id, {})
            self.dyn_reverse_tunnel_rules.setdefault(_id, {})

    def verify_tunnel_rules(self):
        """Make sure that matches & set fields are configured correctly to handle tunnels"""
        if "eth_type" not in self.matches:
            self.matches["eth_type"] = False
        if "in_port" not in self.matches:
            self.matches["in_port"] = False
        if "vlan_vid" not in self.matches:
            self.matches["vlan_vid"] = False
        if "vlan_vid" not in self.set_fields:
            self.set_fields.add("vlan_vid")
        if "vlan_pcp" not in self.matches:
            self.matches["vlan_pcp"] = False
        if "vlan_pcp" not in self.set_fields:
            self.set_fields.add("vlan_pcp")

    def update_reverse_tunnel_rules(
        self, curr_dp, source_id, tunnel_id, out_port, output_table
    ):
        """Update the tunnel rulelist for when the output port has changed (reverse direction)"""
        if not self.requires_reverse_tunnel(tunnel_id):
            return False
        dst_dp = self.tunnel_sources[source_id]["dp"]
        src_dp = self.tunnel_dests[tunnel_id]["dst_dp"]
        prev_list = self.dyn_reverse_tunnel_rules[tunnel_id].get(source_id, [])
        new_list = []
        if curr_dp == src_dp and curr_dp != dst_dp:
            # SRC DP: vlan_vid, vlan_pcp, actions=[out_port]
            # NOTE: For the bi_directional reverse tunnel, we assume that
            #       the packet already has the required encapsulation
            new_list = [{"port": out_port}]
        elif curr_dp == dst_dp and curr_dp != src_dp:
            # DST DP: vlan_vid, vlan_pcp, actions=[pop_vlans, output]
            new_list = [{"pop_vlans": 1}]
            if out_port is None:
                # DP dest tunnel, so we fall through into the eth_dst output table
                new_list.append({"goto": output_table.table_id})
            else:
                # Tunnel has port specified, so output to destination
                new_list.append({"port": out_port})
        elif curr_dp == src_dp and curr_dp == dst_dp:
            # SINGLE DP: actions=[pop_vlans, out_port]
            new_list = [{"pop_vlans": 1}]
            if out_port is None:
                # DP dest tunnel, so we fall through into the eth_dst output table
                new_list.extend([{"goto": output_table.table_id}])
            else:
                # Tunnel has port specified, so output to destination
                new_list.extend([{"port": out_port}])
        else:
            # TRANSIT DP: vlan_vid, vlan_pcp, actions=[output]
            new_list = [{"port": out_port}]
        if new_list != prev_list:
            self.dyn_reverse_tunnel_rules[tunnel_id][source_id] = new_list
            return True
        return True

    def update_source_tunnel_rules(
        self, curr_dp, source_id, tunnel_id, out_port, output_table
    ):
        """Update the tunnel rulelist for when the output port has changed"""
        src_dp = self.tunnel_sources[source_id]["dp"]
        dst_dp = self.tunnel_dests[tunnel_id]["dst_dp"]
        prev_list = self.dyn_tunnel_rules[tunnel_id].get(source_id, [])
        new_list = []
        pcp_flag = valve_of.PCP_TUNNEL_FLAG
        if self.tunnel_dests[tunnel_id]["reverse"]:
            pcp_flag = valve_of.PCP_TUNNEL_REVERSE_DIRECTION_FLAG
        if curr_dp == src_dp and curr_dp != dst_dp:
            # SRC DP: in_port, actions=[push_vlan, output, pop_vlans]
            # Ideally, we would be able to detect if the tunnel has an `allow` action clause.
            #   However, this is difficult as a single ACL can have multiple rules using the same
            #   tunnel, but with one instance requiring the `allow` clause and another, not.
            # This means it is easier to always append the `pop_vlans` in assumption that the
            #   `allow` action does exist, and then optimize/reduce the redundant rules before
            #   outputting the flowrule.
            # We also set the tunnel VLAN header with a PCP value indicating that we are in
            #   the tunnel, which will save the VLANs from being reserved.
            new_list = [
                {"vlan_vids": [{"vid": tunnel_id, "eth_type": ether.ETH_TYPE_8021Q}]},
                {"set_fields": [{"vlan_pcp": pcp_flag}]},
                {"port": out_port},
                {"pop_vlans": 1},
            ]
        elif curr_dp == dst_dp and curr_dp != src_dp:
            # DST DP: in_port, vlan_vid, actions=[pop_vlan, additional_instructions, output]
            # If exit_instructions are applied, then we want to pop off the tunnel
            #   VLAN header, then apply the additional instructions, then output
            if self.tunnel_dests[tunnel_id]["maintain_encapsulation"]:
                # We wish to maintain tunnel encapsulation before outputting
                #   So do not add the pop_vlans rule
                new_list = []
            else:
                new_list = [{"pop_vlans": 1}]
            exit_instructions = self.tunnel_dests[tunnel_id].get(
                "exit_instructions", []
            )
            new_list.extend(copy.copy(list(exit_instructions)))
            if out_port is None:
                # DP dest tunnel, so we fall through into the eth_dst output table
                new_list.append({"goto": output_table.table_id})
            else:
                # Tunnel has port specified, so output to destination
                new_list.append({"port": out_port})
        elif curr_dp == src_dp and curr_dp == dst_dp:
            # SINGLE DP: in_port, actions=[additional_instructions, out_port]
            exit_instructions = self.tunnel_dests[tunnel_id].get(
                "exit_instructions", []
            )
            new_list.extend(copy.copy(list(exit_instructions)))
            if self.tunnel_dests[tunnel_id].get("maintain_encapsulation", False):
                # Maintain encapsulation implies we want the tunnel VID on the packet,
                #   so ensure it is purposefully put onto the packet, even when
                #   there would originally be no need to push on a tunnel VID
                new_list.extend(
                    [
                        {
                            "vlan_vids": [
                                {"vid": tunnel_id, "eth_type": ether.ETH_TYPE_8021Q}
                            ]
                        },
                        {"set_fields": [{"vlan_pcp": pcp_flag}]},
                    ]
                )
            if out_port is None:
                # DP dest tunnel, so we fall through into the eth_dst output table
                new_list.extend([{"goto": output_table.table_id}])
            else:
                # Tunnel has port specified, so output to destination
                new_list.extend([{"port": out_port}])
        else:
            # TRANSIT DP: in_port, vlan_vid, actions=[output]
            new_list = [{"port": out_port}]
        if new_list != prev_list:
            self.dyn_tunnel_rules[tunnel_id][source_id] = new_list
            return True
        return True


# NOTE: 802.1x steals the port ACL table.
PORT_ACL_8021X = ACL(
    "port_acl_8021x",
    0,
    {
        "rules": [
            {
                "eth_type": 1,
                "eth_src": "01:02:03:04:05:06",
                "actions": {
                    "output": {
                        "port": valve_of.ofp.OFPP_LOCAL,
                        "set_fields": [
                            {"eth_src": "01:02:03:04:05:06"},
                            {"eth_dst": "01:02:03:04:05:06"},
                        ],
                    }
                },
            }
        ]
    },
)
PORT_ACL_8021X.build({}, None, 1)

MAB_ACL_8021X = ACL(
    "mab_acl_8021x",
    0,
    {
        "rules": [
            {
                "eth_type": valve_of.ether.ETH_TYPE_IP,
                "eth_src": "01:02:03:04:05:06",
                "ip_proto": valve_of.inet.IPPROTO_UDP,
                "udp_src": 68,
                "udp_dst": 67,
                "actions": {"output": {"port": valve_of.ofp.OFPP_LOCAL}},
            }
        ]
    },
)
MAB_ACL_8021X.build({}, None, 1)
